"""
Some common loss functions (for use with the ``loss`` argument in
`.Simulator.compile`).
"""

import tensorflow as tf


def nan_mse(y_true, y_pred):
    """
    Compute Mean Squared Error between given outputs and targets.

    If any values in ``y_true`` are ``nan``, that will be treated as
    zero error for those elements.

    Parameters
    ----------
    y_true : ``tf.Tensor``
        Target values for a Probe in a network.
    y_pred : ``tf.Tensor``
        Output values from a Probe in a network.

    Returns
    -------
    mse : ``tf.Tensor``
        Tensor representing the mean squared error.
    """

    targets = tf.where(tf.math.is_nan(y_true), y_pred, y_true)
    return tf.reduce_mean(tf.square(targets - y_pred))


class Regularize(tf.losses.Loss):
    """
    An objective function to apply regularization penalties.

    This is designed to be applied to a probed signal, e.g.

    .. testcode::

        with nengo.Network() as net:
            a = nengo.Node([0])
            b = nengo.Node(size_in=1)
            c = nengo.Connection(a, b, transform=1)
            p = nengo.Probe(c, "weights")
            ...

            # this part is optional, but we may often want to only keep the data from
            # the most recent timestep when probing in this way, to save memory
            nengo_dl.configure_settings(keep_history=True)
            net.config[p].keep_history = False

        with nengo_dl.Simulator(net) as sim:
            sim.compile(loss={p: nengo_dl.losses.Regularize()})

    Parameters
    ----------
    order : int or str
        Order of the regularization norm (e.g. ``1`` for L1 norm, ``2`` for
        L2 norm).  See https://www.tensorflow.org/api_docs/python/tf/norm for
        a full description of the possible values for this parameter.
    axis : int or None
        The axis of the probed signal along which to compute norm.  If None
        (the default), the signal is flattened and the norm is computed across
        the resulting vector.  Note that these are only the axes with respect
        to the output on a single timestep (i.e. batch/time dimensions are not
        included).

    Notes
    -----
    The mean will be computed across all the non-``axis`` dimensions after
    computing the norm (including batch/time) in order to compute the overall
    objective value.
    """

    def __init__(self, order=2, axis=None):
        super().__init__()

        self.order = order
        self.axis = axis

    def call(self, y_true, y_pred):
        """
        Invoke the loss function.

        Parameters
        ----------
        y_true : ``tf.Tensor``
            Ignored
        y_pred : ``tf.Tensor``
            The value to be regularized

        Returns
        -------
        output : ``tf.Tensor``
            Scalar regularization loss value.
        """
        if self.axis is None:
            if y_pred.shape.ndims > 3:
                # flatten signal (keeping batch/time dimension)
                y_pred = tf.reshape(
                    y_pred, tf.concat([tf.shape(y_pred)[:2], (-1,)], axis=0)
                )
            axis = 2
        else:
            axis = self.axis + 2

        output = tf.reduce_mean(tf.norm(y_pred, axis=axis, ord=self.order))

        return output



# pylint: disable=missing-docstring

from distutils.version import LooseVersion

import nengo
from nengo.builder.signal import Signal
from nengo.builder.operator import ElementwiseInc, DotInc
import numpy as np
import pkg_resources
import pytest
import tensorflow as tf

import nengo_dl
from nengo_dl.tests import dummies


def test_warn_on_opensim_del(Simulator):
    with nengo.Network() as net:
        nengo.Ensemble(10, 1)

    sim = Simulator(net)
    with pytest.warns(RuntimeWarning):
        sim.__del__()
    sim.close()


def test_args(Simulator):
    class Fn:
        def __init__(self):
            self.last_x = None

        def __call__(self, t, x):
            assert t.dtype == x.dtype
            assert t.shape == ()
            assert isinstance(x, np.ndarray)
            assert self.last_x is not x  # x should be a new copy on each call
            self.last_x = x
            assert np.allclose(x[0], t)

    with nengo.Network() as model:
        u = nengo.Node(lambda t: t)
        v = nengo.Node(Fn(), size_in=1, size_out=0)
        nengo.Connection(u, v, synapse=None)

    with Simulator(model) as sim:
        sim.run(0.01)


def test_signal_init_values(Simulator):
    """Tests that initial values are not overwritten."""

    zero = Signal([0.0])
    one = Signal([1.0])
    five = Signal([5.0])
    zeroarray = Signal([[0.0], [0.0], [0.0]])
    array = Signal([1.0, 2.0, 3.0])

    m = nengo.builder.Model(dt=0)
    m.operators += [ElementwiseInc(zero, zero, five), DotInc(zeroarray, one, array)]

    probes = [
        dummies.Probe(zero, add_to_container=False),
        dummies.Probe(one, add_to_container=False),
        dummies.Probe(five, add_to_container=False),
        dummies.Probe(array, add_to_container=False),
    ]
    m.probes += probes
    for p in probes:
        m.sig[p]["in"] = p.target

    with Simulator(None, model=m) as sim:
        sim.run_steps(3)
        assert np.allclose(sim.data[probes[0]], 0)
        assert np.allclose(sim.data[probes[1]], 1)
        assert np.allclose(sim.data[probes[2]], 5)
        assert np.allclose(sim.data[probes[3]], [1, 2, 3])


def test_entry_point():
    if LooseVersion(tf.__version__) == "1.11.0":
        pytest.xfail("TensorFlow 1.11.0 has conflicting dependencies")

    sims = [
        ep.load(require=False)
        for ep in pkg_resources.iter_entry_points(group="nengo.backends")
    ]
    assert nengo_dl.Simulator in sims


def test_unconnected_node(Simulator):
    hits = np.array(0)
    dt = 0.001

    def f(_):
        hits[...] += 1

    model = nengo.Network()
    with model:
        nengo.Node(f, size_in=0, size_out=0)
    with Simulator(model, unroll_simulation=1) as sim:
        assert hits == 0
        sim.run(dt)
        assert hits == 1
        sim.run(dt)
        assert hits == 2


@pytest.mark.parametrize("bits", ["16", "32", "64"])
def test_dtype(Simulator, request, seed, bits):
    # Ensure dtype is set back to default after the test, even if it fails
    default = nengo.rc.get("precision", "bits")
    request.addfinalizer(lambda: nengo.rc.set("precision", "bits", default))

    float_dtype = np.dtype(getattr(np, "float%s" % bits))
    int_dtype = np.dtype(getattr(np, "int%s" % bits))

    with nengo.Network() as model:
        nengo_dl.configure_settings(dtype="float%s" % bits)

        u = nengo.Node([0.5, -0.4])
        a = nengo.Ensemble(10, 2)
        nengo.Connection(u, a)
        p = nengo.Probe(a)

    with Simulator(model) as sim:
        sim.step()

        # check that the builder has created signals of the correct dtype
        # (note that we may not necessarily use that dtype during simulation)
        for sig in sim.tensor_graph.signals:
            assert sig.dtype in (float_dtype, int_dtype), (
                "Signal '%s' wrong dtype" % sig
            )

        objs = (obj for obj in model.all_objects if sim.data[obj] is not None)
        for obj in objs:
            for x in (x for x in sim.data[obj] if isinstance(x, np.ndarray)):
                assert x.dtype == float_dtype, obj

        assert sim.data[p].dtype == float_dtype


@pytest.mark.parametrize("use_dist", (False, True))
def test_sparse(use_dist, Simulator, rng, seed, monkeypatch):
    # modified version of nengo test_sparse for scipy=False, where we
    # don't expect a warning

    scipy_sparse = pytest.importorskip("scipy.sparse")

    input_d = 4
    output_d = 2
    shape = (output_d, input_d)

    inds = np.asarray([[0, 0], [1, 1], [0, 2], [1, 3]])
    weights = rng.uniform(0.25, 0.75, size=4)
    if use_dist:
        init = nengo.dists.Uniform(0.25, 0.75)
        indices = inds
    else:
        init = scipy_sparse.csr_matrix((weights, inds.T), shape=shape)
        indices = None

    transform = nengo.transforms.Sparse(shape, indices=indices, init=init)

    sim_time = 1.0
    with nengo.Network(seed=seed) as net:
        x = nengo.processes.WhiteSignal(period=sim_time, high=10, seed=seed + 1)
        u = nengo.Node(x, size_out=4)
        a = nengo.Ensemble(100, 2)
        conn = nengo.Connection(u, a, synapse=None, transform=transform)
        ap = nengo.Probe(a, synapse=0.03)

    def run_sim():
        with Simulator(net) as sim:
            sim.run(sim_time)
        return sim

    sim = run_sim()

    actual_weights = sim.data[conn].weights

    full_transform = np.zeros(shape)
    full_transform[inds[:, 0], inds[:, 1]] = weights
    if use_dist:
        actual_weights = actual_weights.toarray()
        assert np.array_equal(actual_weights != 0, full_transform != 0)
        full_transform[:] = actual_weights

    conn.transform = full_transform
    with Simulator(net) as ref_sim:
        ref_sim.run(sim_time)

    assert np.allclose(sim.data[ap], ref_sim.data[ap])


def test_gain_bias(Simulator):
    N = 17
    D = 2

    gain = np.random.uniform(low=0.2, high=5, size=N)
    bias = np.random.uniform(low=0.2, high=1, size=N)

    model = nengo.Network()
    with model:
        a = nengo.Ensemble(N, D)
        a.gain = gain
        a.bias = bias

    with Simulator(model) as sim:
        assert np.allclose(gain, sim.data[a].gain)
        assert np.allclose(bias, sim.data[a].bias)


def test_multirun(Simulator, rng):
    model = nengo.Network(label="Multi-run")

    with Simulator(model) as sim:
        t_stops = sim.dt * rng.randint(low=100, high=2000, size=10)

        # round times to be multiples of 10*dt, so that simulation times will still
        # fall on the right time with unroll_simulation != 1
        t_stops = np.around(t_stops, decimals=2)

        t_sum = 0
        for ti in t_stops:
            sim.run(ti)
            sim_t = sim.trange()
            t = sim.dt * np.arange(1, len(sim_t) + 1)
            assert np.allclose(sim_t, t)

            t_sum += ti
            assert np.allclose(sim_t[-1], t_sum)


def test_time_absolute(Simulator):
    m = nengo.Network()
    with Simulator(m) as sim:
        # modify runtime so that it is a multiple of unroll_simulations
        sim.run(0.01)
    assert np.allclose(sim.trange(), np.arange(sim.dt, 0.01 + sim.dt, sim.dt))


def test_invalid_run_time(Simulator):
    net = nengo.Network()
    with Simulator(net) as sim:
        with pytest.raises(nengo.exceptions.ValidationError):
            sim.run(-0.0001)
        with pytest.warns(UserWarning):
            sim.run(0)
        sim.run(0.0006)  # Rounds up to 0.001

        # may run for more than 1 steps if unroll_simulation != 1
        assert sim.n_steps == sim.unroll


def test_steps(Simulator):
    dt = 0.001
    m = nengo.Network(label="test_steps")
    with Simulator(m, dt=dt) as sim:
        assert sim.n_steps == 0 * sim.unroll
        assert np.allclose(sim.time, 0 * dt * sim.unroll)
        sim.step()
        assert sim.n_steps == 1 * sim.unroll
        assert np.allclose(sim.time, 1 * dt * sim.unroll)
        sim.step()
        assert sim.n_steps == 2 * sim.unroll
        assert np.allclose(sim.time, 2 * dt * sim.unroll)

        assert np.isscalar(sim.n_steps)
        assert np.isscalar(sim.time)



"""
Manages the data and build processes associated with implementing a Nengo simulation
in TensorFlow.
"""

from collections import OrderedDict, defaultdict
import logging
import warnings

from nengo import Connection, Process
from nengo.builder.neurons import SimNeurons
from nengo.builder.operator import Reset, SimPyFunc, TimeUpdate
from nengo.builder.processes import SimProcess
from nengo.config import ConfigError
from nengo.exceptions import BuildError
from nengo.neurons import Direct
from nengo.synapses import Lowpass
from nengo.transforms import SparseMatrix
import numpy as np
import tensorflow as tf
from tensorflow.python.eager import context
from tensorflow.python.training.tracking import base as trackable

from nengo_dl import (
    builder,
    config,
    compat,
    graph_optimizer,
    tensor_node,
    signals,
    utils,
)

logger = logging.getLogger(__name__)


class TensorGraph(tf.keras.layers.Layer):
    """
    Implement the Nengo simulation as a Keras Layer.

    Parameters
    ----------
    model : `~nengo.builder.Model`
        Pre-built Nengo model describing the network to be simulated.
    dt : float
        Length of a simulator timestep, in seconds.
    unroll_simulation : int
        Unroll simulation loop by explicitly building ``unroll_simulation``
        iterations into the computation graph.
    minibatch_size : int
        The number of simultaneous inputs that will be passed through the
        network.
    device : None or ``"/cpu:0"`` or ``"/gpu:[0-n]"``
        Device on which to execute computations (if None then uses the
        default device as determined by TensorFlow).
    progress : `.utils.ProgressBar`
        Progress bar for optimization stage.
    seed : int
        Seed for random number generation.
    """

    @trackable.no_automatic_dependency_tracking
    def __init__(
        self, model, dt, unroll_simulation, minibatch_size, device, progress, seed
    ):
        super().__init__(
            name="TensorGraph",
            dynamic=False,
            trainable=not config.get_setting(model, "inference_only", False),
            dtype=config.get_setting(model, "dtype", "float32"),
            batch_size=minibatch_size,
        )

        self.model = model
        self.dt = dt
        self.unroll = unroll_simulation
        self.use_loop = config.get_setting(model, "use_loop", True)
        self.minibatch_size = minibatch_size
        self.device = device
        self.seed = seed
        self.inference_only = not self.trainable
        self.signals = signals.SignalDict(self.dtype, self.minibatch_size)

        # find invariant inputs (nodes that don't receive any input other
        # than the simulation time). we'll compute these outside the simulation
        # and feed in the result.
        if self.model.toplevel is None:
            self.invariant_inputs = OrderedDict()
        else:
            self.invariant_inputs = OrderedDict(
                (n, n.output)
                for n in self.model.toplevel.all_nodes
                if n.size_in == 0 and not isinstance(n, tensor_node.TensorNode)
            )

        # remove input nodes because they are executed outside the simulation
        node_processes = [
            n.output for n in self.invariant_inputs if isinstance(n.output, Process)
        ]
        operators = [
            op
            for op in self.model.operators
            if not (
                (isinstance(op, SimPyFunc) and op.x is None)
                or (
                    isinstance(op, SimProcess)
                    and op.input is None
                    and op.process in node_processes
                )
            )
        ]

        # mark trainable signals
        self.mark_signals()

        logger.info("Initial plan length: %d", len(operators))

        # apply graph simplification functions
        simplifications = config.get_setting(
            model, "simplifications", graph_optimizer.default_simplifications,
        )

        with progress.sub("operator simplificaton", max_value=None):
            old_operators = []
            while len(old_operators) != len(operators) or any(
                x is not y for x, y in zip(operators, old_operators)
            ):
                old_operators = operators
                for simp in simplifications:
                    operators = simp(operators)

        # group mergeable operators
        planner = config.get_setting(model, "planner", graph_optimizer.tree_planner)

        with progress.sub("merging operators", max_value=None):
            plan = planner(operators)

        # TODO: we could also merge operators sequentially (e.g., combine
        # a copy and dotinc into one op), as long as the intermediate signal
        # is only written to by one op and read by one op

        # order signals/operators to promote contiguous reads
        sorter = config.get_setting(model, "sorter", graph_optimizer.order_signals)

        with progress.sub("ordering signals", max_value=None):
            sigs, self.plan = sorter(plan, n_passes=10)

        # create base arrays and map Signals to TensorSignals (views on those
        # base arrays)
        with progress.sub("creating signals", max_value=None):
            self.create_signals(sigs)

        # generate unique names for layer inputs/outputs
        # this follows the TensorFlow unique naming scheme, so if multiple objects are
        # created with the same name, they will be named like name, NAME_1, name_2
        # (note: case insensitive)
        self.io_names = {}
        name_count = defaultdict(int)
        for obj in list(self.invariant_inputs.keys()) + self.model.probes:
            name = (
                type(obj).__name__.lower()
                if obj.label is None
                else utils.sanitize_name(obj.label)
            )

            key = name.lower()

            if name_count[key] > 0:
                name += "_%d" % name_count[key]

            self.io_names[obj] = name
            name_count[key] += 1

        logger.info("Optimized plan length: %d", len(self.plan))
        logger.info(
            "Number of base arrays: (%s, %d), (%s, %d), (%s, %d)",
            *tuple((k, len(x)) for k, x in self.base_arrays_init.items()),
        )

    def build_inputs(self):
        """
        Generates a set of Input layers that can be used as inputs to a
        TensorGraph layer.

        Returns
        -------
        n_steps : ``tf.keras.layers.Input``
            Input layer for specifying the number of simulation timesteps.
        inputs : dict of {`nengo.Node`: ``tf.keras.layers.Input``}
            Input layers for each of the Nodes in the network.
        """

        # input placeholders
        inputs = OrderedDict()
        for n in self.invariant_inputs:
            inputs[n] = tf.keras.layers.Input(
                shape=(None, n.size_out),
                batch_size=self.minibatch_size,
                dtype=self.dtype,
                name=self.io_names[n],
            )

        # number of steps to run
        n_steps = tf.keras.layers.Input(
            shape=(1,), batch_size=self.minibatch_size, dtype="int32", name="n_steps"
        )

        return inputs, n_steps

    def build(self, input_shape=None):
        """
        Create any Variables used in the model.

        Parameters
        ----------
        input_shape : list of tuple of int
            Shapes of all the inputs to this layer.
        """

        super().build(input_shape)

        tf.random.set_seed(self.seed)

        def get_initializer(init_vals):
            """Use more efficient initializers if possible to save memory."""

            values, shapes, dtype, minibatched = init_vals

            # initial value of None means that the initial value isn't used, so we
            # can use anything for the initial value
            if all(v is None for v in values):
                initializer = None
            elif all(v is None or np.all(v == 0) for v in values):
                initializer = tf.initializers.zeros()
            elif all(v is None or np.all(v == 1) for v in values):
                initializer = tf.initializers.ones()
            else:
                val = tf.concat(
                    [
                        tf.zeros(s, dtype)
                        if v is None
                        else tf.cast(tf.broadcast_to(v, s), dtype)
                        for v, s in zip(values, shapes)
                    ],
                    axis=1 if minibatched else 0,
                )
                initializer = lambda shape=None, dtype=None: val

            # figure out shape of full concatenated initial value
            shape = list(shapes[0])
            shape[minibatched] = sum(x[minibatched] for x in shapes)

            return initializer, tuple(shape), dtype

        # variables for model parameters
        with trackable.no_automatic_dependency_tracking_scope(self):
            self.base_params = OrderedDict()
        assert len(self.base_params) == 0
        for sig_type in ("trainable", "non_trainable"):
            for k, v in self.base_arrays_init[sig_type].items():
                initializer, shape, dtype = get_initializer(v)
                assert initializer is not None  # params should never be set
                self.base_params[k] = self.add_weight(
                    initializer=initializer,
                    shape=shape,
                    dtype=dtype,
                    trainable=sig_type == "trainable",
                    name="base_params/%s_%s_%s"
                    % (sig_type, dtype, "_".join(str(x) for x in shape)),
                )

        logger.debug("created base param variables")
        logger.debug([str(x) for x in self.base_params.values()])

        # variables to save the internal state of simulation between runs
        with trackable.no_automatic_dependency_tracking_scope(self):
            self.saved_state = OrderedDict()
        for k, v in self.base_arrays_init["state"].items():
            initializer, shape, dtype = get_initializer(v)
            if initializer is not None:
                # don't need to save the state for signals where the initial value
                # doesn't matter
                self.saved_state[k] = tf.Variable(
                    initial_value=lambda: initializer(shape=shape, dtype=dtype),
                    shape=shape,
                    dtype=dtype,
                    trainable=False,
                    name="saved_state/%s_%s" % (dtype, "_".join(str(x) for x in shape)),
                )

        logger.debug("created saved state variables")
        logger.debug([str(x) for x in self.saved_state.values()])

        # call build on any TensorNode Layers

        def unbuild(layer):
            assert layer.built

            # clear any losses attached to layer (they will be recreated in the
            # build step, so we don't want to keep around any losses
            # associated with the previous build)
            # note: not clearing layer._losses, because those are manually added
            # by the user (not created during the build process)
            layer._eager_losses = []
            layer._callable_losses = []

            layer.built = False

            for sub in layer._layers:
                if isinstance(sub, tf.keras.layers.Layer):
                    unbuild(sub)

        layer_ops = [
            op
            for ops in self.plan
            if isinstance(ops[0], tensor_node.SimTensorNode)
            for op in ops
            if isinstance(op.func, tf.keras.layers.Layer)
        ]
        weight_gets = []
        weight_sets = []
        for op in layer_ops:
            if op.func in self._layers:
                # already built this layer
                continue

            if op.time is None:
                shape_in = []
            else:
                shape_in = [()]
            if op.input is not None:
                shape_in += [(self.minibatch_size,) + op.shape_in]
            if len(shape_in) == 1:
                shape_in = shape_in[0]

            if op.func.built:
                # we rebuild the layer (even if it is already built),
                # because we need to build the weights within the TensorGraph
                # context

                # save the weight values so they can be restored
                # exactly inside the tensornode
                weights = op.func.weights
                weight_gets.extend(weights)

                # clear the results of previous build
                unbuild(op.func)
            else:
                weights = None

            with tf.name_scope(op.func.name):
                op.func.build(shape_in)

            if weights is not None:
                weight_sets.extend(op.func.weights)

            # add op func to _layers so that any weights are collected
            self._layers.append(op.func)

        if len(weight_gets) > 0:
            # do all the weight getting/setting in one go, for efficiency reasons
            ctx = (
                weight_gets[0].graph.as_default()
                if hasattr(weight_gets[0], "graph")
                else context.eager_mode()
            )
            with ctx:
                weight_vals = tf.keras.backend.batch_get_value(weight_gets)

            tf.keras.backend.batch_set_value(zip(weight_sets, weight_vals))

        # initialize state variables (need to do this manually because we're not
        # adding them to self.weights)
        # note: don't need to do this in eager mode, since variables are
        # initialized on creation
        # TODO: why does this cause problems if it is done before the tensornode
        #  weight get/sets above?
        if not context.executing_eagerly():
            tf.keras.backend.batch_get_value(
                [var.initializer for var in self.saved_state.values()]
            )

    # @tf.function  # TODO: get this working? does this help?
    @tf.autograph.experimental.do_not_convert
    def call(self, inputs, training=None, progress=None, stateful=False):
        """
        Constructs the graph elements to simulate the model.

        Parameters
        ----------
        inputs : list of ``tf.Tensor``
            Input layers/tensors for the network (must match the structure defined in
            `.build_inputs`).
        training : bool
            Whether the network is being run in training or inference mode.  If None,
            uses the symbolic Keras learning phase variable.
        progress : `.utils.ProgressBar`
            Progress bar for construction stage.
        stateful : bool
            Whether or not to build the model to support preserving the internal state
            between executions.

        Returns
        -------
        probe_arrays : list of ``tf.Tensor``
            Tensors representing the output of all the Probes in the network (order
            corresponding to ``self.model.probes``, which is the order the Probes were
            instantiated).
        """

        super().call(inputs, training=training)

        if training == 1 and self.inference_only:
            raise BuildError(
                "TensorGraph was created with inference_only=True; cannot be built "
                "with training=%s" % training
            )

        tf.random.set_seed(self.seed)

        if progress is None:
            progress = utils.NullProgressBar()

        # reset signaldict
        self.signals.reset()

        # create these constants once here for reuse in different operators
        self.signals.dt = tf.constant(self.dt, self.dtype)
        self.signals.dt_val = self.dt  # store the actual value as well
        self.signals.zero = tf.constant(0, self.dtype)
        self.signals.one = tf.constant(1, self.dtype)

        # set up invariant inputs
        with trackable.no_automatic_dependency_tracking_scope(self):
            self.node_inputs = {}
        for n, inp in zip(self.invariant_inputs, inputs):
            # specify shape of inputs (keras sometimes loses this shape information)
            inp.set_shape([self.minibatch_size, inp.shape[1], n.size_out])

            self.node_inputs[n] = inp

        self.steps_to_run = inputs[-1][0, 0]

        # initialize op builder
        build_config = builder.BuildConfig(
            inference_only=self.inference_only,
            lif_smoothing=config.get_setting(self.model, "lif_smoothing"),
            cpu_only=self.device == "/cpu:0" or not utils.tf_gpu_installed,
            rng=np.random.RandomState(self.seed),
            training=(
                tf.keras.backend.learning_phase() if training is None else training
            ),
        )
        self.op_builder = builder.Builder(self.plan, self.signals, build_config)

        # pre-build stage
        with progress.sub("pre-build stage", max_value=len(self.plan)) as sub:
            self.op_builder.build_pre(sub)

        # build stage
        with progress.sub("build stage", max_value=len(self.plan) * self.unroll) as sub:
            steps_run, probe_arrays, final_internal_state, final_base_params = (
                self._build_loop(sub) if self.use_loop else self._build_no_loop(sub)
            )

        # store these so that they can be accessed after the initial build
        with trackable.no_automatic_dependency_tracking_scope(self):
            self.steps_run = steps_run
            self.probe_arrays = probe_arrays
            self.final_internal_state = final_internal_state
            self.final_base_params = final_base_params

        # logging
        logger.info(
            "Number of reads: %d", sum(x for x in self.signals.read_types.values())
        )
        for x in self.signals.read_types.items():
            logger.info("    %s: %d", *x)
        logger.info(
            "Number of writes: %d", sum(x for x in self.signals.write_types.values())
        )
        for x in self.signals.write_types.items():
            logger.info("    %s: %d", *x)

        # note: always return steps_run so that the simulation will run for the given
        # number of steps, even if there are no output probes
        outputs = list(probe_arrays.values()) + [steps_run]

        updates = []
        if stateful:
            # update saved state
            updates.extend(
                var.assign(val)
                for var, val in zip(self.saved_state.values(), final_internal_state)
            )

        # if any of the base params have changed (due to online learning rules) then we
        # also need to assign those back to the original variable (so that their
        # values will persist). any parameters targeted by online learning rules
        # will be minibatched, so we only need to update the minibatched params.
        for (key, var), val in zip(self.base_params.items(), final_base_params):
            try:
                minibatched = self.base_arrays_init["non_trainable"][key][-1]
            except KeyError:
                minibatched = self.base_arrays_init["trainable"][key][-1]

            if minibatched:
                updates.append(var.assign(val))

        logger.info("Number of variable updates: %d", len(updates))

        if len(updates) > 0:
            with tf.control_dependencies(updates):
                outputs = [tf.identity(x) for x in outputs]

        return outputs

    def _fill_bases(self, saved_state, base_params):
        """
        Initialize signals.bases from TensorGraph params.

        Parameters
        ----------
        saved_state : dict
            Mapping from base keys to initial values
        base_params : dict
            Mapping from base keys to initial values
        """

        for key, val in saved_state.items():
            # we add the tf.identity so that when we write we're not updating
            # the base variable
            self.signals.bases[key] = tf.identity(val)
        for key, val in base_params.items():
            self.signals.bases[key] = tf.identity(val)
        for key, (_, shapes, _, minibatched) in self.base_arrays_init["state"].items():
            if key not in self.signals.bases:
                # no saved state for this base, so we just temporarily insert
                # the shape information so that future scatters will know
                # what the base shape is
                shape = list(shapes[0])
                shape[minibatched] = sum(x[minibatched] for x in shapes)
                self.signals.bases[key] = tuple(shape)

    def _build_loop(self, progress):
        """
        Build simulation loop using symbolic while loop.

        Parameters
        ----------
        progress : `.utils.ProgressBar`
            Progress bar for loop construction

        Returns
        -------
        steps_run : ``tf.Tensor``
            The number of simulation steps that were executed.
        probe_arrays : dict of {`nengo.Probe`: ``tf.Tensor``}
            Arrays containing the output values for each Probe.
        final_internal_state: list of ``tf.Tensor``
            Tensors representing the value of all internal state at the end of the run.
        """

        def loop_condition(loop_i, n_steps, *_):
            return loop_i < n_steps

        def loop_body(loop_i, n_steps, probe_arrays, saved_state, base_params):
            # fill in signals.bases
            # note: we need to do this here because we
            # need to use the tensors from inside the loop, not the source variables)
            self._fill_bases(
                dict(zip(self.saved_state, saved_state)),
                dict(zip(self.base_params, base_params)),
            )

            def update_probes(probe_tensors, loop_i):
                for i, p in enumerate(probe_tensors):
                    if config.get_setting(
                        self.model,
                        "keep_history",
                        default=True,
                        obj=self.model.probes[i],
                    ):
                        probe_arrays[i] = probe_arrays[i].write(loop_i, p)
                    else:
                        probe_arrays[i] = tf.cond(
                            pred=tf.equal(loop_i + 1, n_steps),
                            true_fn=lambda p=p, i=i: probe_arrays[i].write(0, p),
                            false_fn=lambda i=i: probe_arrays[i],
                        )

            loop_i = self._build_inner_loop(loop_i, update_probes, progress)

            state_arrays = tuple(self.signals.bases[key] for key in self.saved_state)
            base_arrays = tuple(self.signals.bases[key] for key in self.base_params)

            return loop_i, n_steps, probe_arrays, state_arrays, base_arrays

        loop_i = tf.constant(0)

        probe_arrays = [
            tf.TensorArray(self.dtype, clear_after_read=True, size=0, dynamic_size=True)
            for _ in self.model.probes
        ]

        # build simulation loop
        loop_vars = (
            loop_i,
            self.steps_to_run,
            probe_arrays,
            tuple(self.saved_state.values()),
            tuple(self.base_params.values()),
        )

        loop_vars = tf.while_loop(
            cond=loop_condition,
            body=loop_body,
            loop_vars=loop_vars,
            parallel_iterations=1,  # TODO: parallel iterations work in eager mode
        )

        # change to shape (minibatch_size,) (required by keras) instead of a scalar
        steps_run = tf.tile(tf.expand_dims(loop_vars[0], 0), (self.minibatch_size,))

        probe_arrays = OrderedDict()
        for p, a in zip(self.model.probes, loop_vars[2]):
            x = a.stack()

            if self.model.sig[p]["in"].minibatched:
                # change from tensorarray's (steps, batch, d) to (batch, steps, d)
                perm = np.arange(x.shape.ndims)
                perm[[0, 1]] = perm[[1, 0]]
                x = tf.transpose(x, perm=perm)
            else:
                # add minibatch dimension for consistency
                x = tf.expand_dims(x, 0)

            probe_arrays[p] = x

        final_internal_state = loop_vars[3]
        final_base_params = loop_vars[4]

        return steps_run, probe_arrays, final_internal_state, final_base_params

    def _build_no_loop(self, progress):
        """
        Build simulation loop through explicit unrolling.

        Parameters
        ----------
        progress : `.utils.ProgressBar`
            Progress bar for loop construction

        Returns
        -------
        steps_run : ``tf.Tensor``
            The number of simulation steps that were executed.
        probe_arrays : dict of {`nengo.Probe`: ``tf.Tensor``}
            Arrays containing the output values for each Probe.
        final_internal_state: list of ``tf.Tensor``
            Tensors representing the value of all internal state at the end of the run.
        """

        self._fill_bases(self.saved_state, self.base_params)

        loop_i = tf.constant(0)  # symbolic loop variable
        loop_iter = 0  # non-symbolic loop variable
        probe_data = [[] for _ in self.model.probes]

        def update_probes(probe_tensors, _):
            nonlocal loop_iter

            for i, p in enumerate(probe_tensors):
                if config.get_setting(
                    self.model, "keep_history", default=True, obj=self.model.probes[i]
                ):
                    probe_data[i].append(p)
                elif loop_iter == self.unroll - 1:
                    probe_data[i].append(p)

            loop_iter += 1

        loop_i = self._build_inner_loop(loop_i, update_probes, progress)

        # change to shape (minibatch_size,) (required by keras) instead of a scalar
        steps_run = tf.tile(tf.expand_dims(loop_i, 0), (self.minibatch_size,))

        probe_arrays = OrderedDict()
        for p, a in zip(self.model.probes, probe_data):
            if self.model.sig[p]["in"].minibatched:
                x = tf.stack(a, axis=1)
            else:
                x = tf.stack(a, axis=0)

                # add minibatch dimension for consistency
                x = tf.expand_dims(x, 0)

            probe_arrays[p] = x

        final_internal_state = tuple(
            self.signals.bases[key] for key in self.saved_state
        )
        final_base_params = tuple(self.signals.bases[key] for key in self.base_params)

        return steps_run, probe_arrays, final_internal_state, final_base_params

    def _build_inner_loop(self, loop_i, update_probes, progress):
        """

        Parameters
        ----------
        loop_i : ``tf.Tensor``
            Loop iteration variable.
        update_probes : callable
            Function that will update some stored probe data in each iteration.
        progress
            Progress bar for loop construction.

        Returns
        -------
        loop_i : ``tf.Tensor``
            Updated loop iteration variable.
        """

        constant_probes = {}
        for p in self.model.probes:
            probe_sig = self.model.sig[p]["in"]
            if probe_sig not in self.signals:
                # if a probe signal isn't in sig_map, that means that it
                # isn't involved in any simulator ops.  so we know its value
                # never changes, and we'll just return a constant containing
                # the initial value.
                init_val = probe_sig.initial_value
                if probe_sig.minibatched:
                    init_val = np.tile(init_val[None, :], (self.minibatch_size, 1))

                constant_probes[p] = tf.constant(init_val, dtype=self.dtype)

        for unroll_iter in range(self.unroll):
            logger.debug("BUILDING ITERATION %d", unroll_iter)
            with tf.name_scope("iteration_%d" % unroll_iter):
                # fill in invariant input data
                for n in self.node_inputs:
                    if self.model.sig[n]["out"] in self.signals:
                        # if the out signal doesn't exist then that means that
                        # the node output isn't actually used anywhere, so we can
                        # ignore it

                        self.signals.scatter(
                            self.signals[self.model.sig[n]["out"]],
                            self.node_inputs[n][:, loop_i],
                        )

                # build the operators for a single step
                # note: we tie things to the `loop_i` variable so that we
                # can be sure the other things we're tying to the
                # simulation step (side effects and probes) from the
                # previous timestep are executed before the next step
                # starts
                with tf.control_dependencies([loop_i]):
                    # build operators
                    side_effects = self.op_builder.build(progress)

                    logger.debug("collecting probe tensors")
                    probe_tensors = []
                    for p in self.model.probes:
                        if p in constant_probes:
                            probe_tensors.append(constant_probes[p])
                        else:
                            probe_tensors.append(
                                self.signals.gather(
                                    self.signals[self.model.sig[p]["in"]]
                                )
                            )

                    logger.debug("=" * 30)
                    logger.debug("build_step complete")
                    logger.debug("probe_tensors %s", [str(x) for x in probe_tensors])
                    logger.debug("side_effects %s", [str(x) for x in side_effects])

                # update probe data
                update_probes(probe_tensors, loop_i)

                # need to make sure that any operators that could have side
                # effects run each timestep, so we tie them to the loop
                # increment. we also need to make sure that all the probe
                # reads happen before those values get overwritten on the
                # next timestep
                with tf.control_dependencies(side_effects + probe_tensors):
                    loop_i += 1

        return loop_i

    @trackable.no_automatic_dependency_tracking
    def build_post(self):
        """
        Executes post-build processes for operators (after the graph has
        been constructed and whenever Simulator is reset).
        """

        rng = np.random.RandomState(self.seed)

        # build input functions (we need to do this here, because in the case
        # of processes these functions need to be be rebuilt on reset)
        self.input_funcs = {}
        for n, output in self.invariant_inputs.items():
            if isinstance(output, np.ndarray):
                self.input_funcs[n] = output
            elif isinstance(output, Process):
                state = output.make_state((n.size_in,), (n.size_out,), self.dt)
                self.input_funcs[n] = [
                    output.make_step(
                        (n.size_in,),
                        (n.size_out,),
                        self.dt,
                        output.get_rng(rng),
                        state,
                    )
                    for _ in range(self.minibatch_size)
                ]
            elif n.size_out > 0:
                self.input_funcs[n] = [
                    utils.align_func((n.size_out,), self.dtype)(output)
                ]
            else:
                # a node with no inputs and no outputs, but it can still
                # have side effects
                self.input_funcs[n] = [output]

        # execute build_post on all the op builders
        self.op_builder.build_post()

    def get_tensor(self, sig):
        """
        Returns a Tensor corresponding to the given Signal.

        Parameters
        ----------
        sig : `~nengo.builder.Signal`
            A signal in the Nengo model.

        Returns
        -------
        tensor : ``tf.Tensor``
            Tensor containing the value of the given Signal.
        """

        tensor_sig = self.signals[sig]

        try:
            base = self.base_params[tensor_sig.key]
        except KeyError:
            base = self.saved_state[tensor_sig.key]

        return tf.gather(
            base, tensor_sig.tf_indices, axis=1 if tensor_sig.minibatched else 0,
        )

    def mark_signals(self):
        """
        Mark all the signals in ``self.model`` according to whether they
        represent trainable parameters of the model (parameters that can be
        optimized by deep learning methods).

        Trainable parameters include connection weights, ensemble encoders, and
        neuron biases.  Unless one of those signals is targeted by a Nengo
        learning rule (otherwise the learning rule update conflicts with the
        deep learning optimization).

        Users can manually specify whether signals are trainable or not using
        the config system (e.g.,
        ``net.config[nengo.Ensemble].trainable = False``).

        The trainable attribute will be set to one of three values:

        - ``True``: Signal is trainable
        - ``False``: Signal could be trainable, but has been set to non-trainable
          (e.g., because the user manually configured that object not to be trainable).
        - ``None``: Signal is never trainable (e.g., simulator state)
        """

        def get_trainable(parent_configs, obj):
            """Looks up the current value of ``obj.trainable``."""

            if self.inference_only:
                return False

            # default to 1 (so that we can distinguish between an object being
            # set to trainable vs defaulting to trainable)
            trainable = 1

            # we go from top down (so lower level settings will override)
            for cfg in parent_configs:
                try:
                    cfg_trainable = getattr(cfg[obj], "trainable", None)
                except ConfigError:
                    # object not configured in this network config
                    cfg_trainable = None

                if cfg_trainable is not None:
                    trainable = cfg_trainable

            return trainable

        def mark_network(parent_configs, net):
            """Recursively marks the signals for objects within each subnetwork."""

            parent_configs = parent_configs + [net.config]

            for subnet in net.networks:
                mark_network(parent_configs, subnet)

            # encoders and biases are trainable
            for ens in net.ensembles:
                ens_trainable = get_trainable(parent_configs, ens)

                self.model.sig[ens]["encoders"].trainable = ens_trainable
                self.model.sig[ens]["encoders"].minibatched = False

                if not isinstance(ens.neuron_type, Direct):
                    neurons_trainable = get_trainable(parent_configs, ens.neurons)
                    if neurons_trainable is 1:  # noqa: F632
                        neurons_trainable = ens_trainable

                    self.model.sig[ens.neurons]["bias"].trainable = neurons_trainable
                    self.model.sig[ens.neurons]["bias"].minibatched = False

            # connection weights are trainable
            for conn in net.connections:
                # note: this doesn't include probe connections, since they
                # aren't added to the network
                if compat.conn_has_weights(conn):
                    self.model.sig[conn]["weights"].trainable = get_trainable(
                        parent_configs, conn
                    )
                    self.model.sig[conn]["weights"].minibatched = False

            # parameters can't be modified by an online Nengo learning rule
            # and offline training at the same time. (it is possible in
            # theory, but it complicates things a lot and is probably not a
            # common use case). we also make those signals minibatched
            # (they wouldn't be normally), because we want to be able to
            # learn independently in each minibatch
            for conn in net.connections:
                rule = conn.learning_rule
                if rule is not None:
                    if isinstance(rule, dict):
                        rule = list(rule.values())
                    elif not isinstance(rule, list):
                        rule = [rule]

                    for r in rule:
                        if r.modifies in ("weights", "decoders"):
                            obj = conn
                            attr = "weights"
                        elif r.modifies == "encoders":
                            obj = conn.post_obj
                            attr = "encoders"
                        else:
                            raise NotImplementedError

                        if self.model.sig[obj][attr].trainable is True:
                            warnings.warn(
                                "%s has a learning rule and is also set "
                                "to be trainable; this is likely to "
                                "produce strange training behaviour." % obj
                            )
                        else:
                            self.model.sig[obj][attr].trainable = False

                        self.model.sig[obj][attr].minibatched = True

        if self.model.toplevel is None:
            warnings.warn(
                "No top-level network in model; assuming no trainable parameters",
                UserWarning,
            )
        else:
            mark_network([], self.model.toplevel)

            # the connections to connection probes are not trainable, but
            # also not minibatched
            probe_seeds = [self.model.seeds[p] for p in self.model.probes]
            for obj, seed in self.model.seeds.items():
                if isinstance(obj, Connection) and seed in probe_seeds:
                    if compat.conn_has_weights(obj):
                        self.model.sig[obj]["weights"].trainable = None
                        self.model.sig[obj]["weights"].minibatched = False

        # time/step are not minibatched and not trainable
        self.model.step.trainable = None
        self.model.step.minibatched = False
        self.model.time.trainable = None
        self.model.time.minibatched = False

        # fill in defaults for all other signals
        # signals are not trainable by default, and views take on the
        # properties of their bases
        for op in self.model.operators:
            for sig in op.all_signals:
                if not hasattr(sig.base, "trainable"):
                    sig.base.trainable = None

                if not hasattr(sig.base, "minibatched"):
                    sig.base.minibatched = not sig.base.trainable

                if not hasattr(sig, "trainable"):
                    sig.trainable = sig.base.trainable

                if not hasattr(sig, "minibatched"):
                    sig.minibatched = sig.base.minibatched

    @trackable.no_automatic_dependency_tracking
    def create_signals(self, sigs):
        """
        Groups signal data together into larger arrays, and represent each
        individual signal as a slice into that array.

        Parameters
        ----------
        sigs : list of `~nengo.builder.Signal`
            Base signals arranged into the order in which they should reside in
            memory (e.g., output from `.graph_optimizer.order_signals`)
        """

        base_arrays = OrderedDict(
            [
                ("trainable", OrderedDict()),
                ("non_trainable", OrderedDict()),
                ("state", OrderedDict()),
            ]
        )
        curr_keys = {}
        sig_idxs = {s: i for i, s in enumerate(sigs)}

        # find the non-overlapping partitions of the signals
        breaks = []
        diff = defaultdict(int)
        for ops in self.plan:
            if isinstance(ops[0], Reset):
                # don't include Resets, otherwise the big reset block
                # overrides most of the partitioning
                partition_sigs = []
            else:
                partition_sigs = range(len(ops[0].all_signals))

            for i in partition_sigs:
                op_sigs = [op.all_signals[i].base for op in ops]
                idxs = [sig_idxs[s] for s in op_sigs]
                diff[op_sigs[np.argmin(idxs)]] += 1
                diff[op_sigs[np.argmax(idxs)]] -= 1

        # find the partition points in signal list
        open = 0
        for i, s in enumerate(sigs):
            if s in diff:
                open += diff[s]

            if open == 0:
                breaks += [i + 1]

        logging.debug("partitions")
        logging.debug(
            "\n%s", "".join("|" if i in breaks else " " for i in range(len(sigs)))
        )

        # find all the signals that have a set operation associated with them

        def special_set(s, op):
            return (
                # we don't include Lowpass ops, because for efficiency reasons in the
                # nengo-dl Lowpass implementation we reuse the output signal (which is
                # set) as the state signal (so we need to include that signal in the
                # state)
                (isinstance(op, SimProcess) and isinstance(op.process, Lowpass))
                # nengo marks the time step as a set, but really it's an inc (since
                # it's incrementing the simulation step)
                or (isinstance(op, TimeUpdate) and s is op.step)
                # nengo marks neuron state as a set, but really it's more like an
                # inc/update (since the neuron calculation may depend on the state)
                or (isinstance(op, SimNeurons) and s in op.states)
            )

        set_sigs = {
            s.base
            for ops in self.plan
            for op in ops
            for s in op.sets
            if not special_set(s, op)
        }

        # create all the base signals
        for i, sig in enumerate(sigs):
            assert sig not in self.signals
            assert not sig.is_view

            if i in breaks:
                # start a new array for all current bases
                for k in curr_keys:
                    curr_keys[k] = object()

            # convert to appropriate dtype
            if np.issubdtype(sig.dtype, np.floating):
                dtype = self.dtype
            elif np.issubdtype(sig.dtype, np.integer):
                dtype = "int32"
            elif np.issubdtype(sig.dtype, np.bool_):
                dtype = "bool"
            else:
                raise NotImplementedError("Unsupported signal dtype")

            if sig.sparse:
                # for sparse tensors, what we care about is the shape of the
                # underlying data, not the full matrix
                shape = (sig.initial_value.size,)
            else:
                # resize scalars to length 1 vectors
                shape = sig.shape if sig.shape != () else (1,)

            # parameters of signal that affect the base array
            array_params = (dtype, shape[1:], sig.trainable, sig.minibatched)

            # key used to map signals to base arrays
            if array_params not in curr_keys:
                curr_keys[array_params] = object()
            key = curr_keys[array_params]

            if sig in set_sigs:
                # signals with a set operation associated with them don't need an
                # initial value (since the value will just be immediately overridden
                # by the set operation)
                initial_value = None
            else:
                initial_value = sig.initial_value
                if sig.sparse:
                    if isinstance(initial_value, SparseMatrix):
                        initial_value = initial_value.data
                    else:
                        initial_value = initial_value.tocoo().data

            if sig.minibatched:
                shape = (self.minibatch_size,) + shape

            if sig.trainable is None:
                sig_type = "state"
            elif sig.trainable:
                sig_type = "trainable"
            else:
                sig_type = "non_trainable"

            if key in base_arrays[sig_type]:
                base_arrays[sig_type][key][0].append(initial_value)
                base_arrays[sig_type][key][1].append(shape)
            else:
                base_arrays[sig_type][key] = [
                    [initial_value],
                    [shape],
                    dtype,
                    sig.minibatched,
                ]

            n = sum(x[sig.minibatched] for x in base_arrays[sig_type][key][1])
            slices = [(n - shape[sig.minibatched], n)]

            tensor_sig = self.signals.get_tensor_signal(
                slices,
                key,
                dtype,
                shape[sig.minibatched :],
                sig.minibatched,
                label=sig.name,
                signal=sig,
            )

            logger.debug("created base signal")
            logger.debug(sig)
            logger.debug(tensor_sig)

        # add any signal views to the sig_map
        all_views = [
            sig
            for ops in self.plan
            for op in ops
            for sig in op.all_signals
            if sig.is_view
        ]
        for sig in all_views:
            if sig.size == sig.base.size:
                # reshape view
                self.signals[sig] = self.signals[sig.base].reshape(sig.shape)
            else:
                if sig.shape[1:] != sig.base.shape[1:]:
                    # TODO: support this?
                    raise NotImplementedError("Slicing on axes > 0 is not supported")

                # slice view
                assert np.all([x == 1 for x in sig.elemstrides[1:]])

                start = sig.elemoffset
                stride = sig.elemstrides[0]
                stop = start + sig.size * stride
                if stop < 0:
                    stop = None

                self.signals[sig] = self.signals[sig.base][slice(start, stop, stride)]

        self.base_arrays_init = base_arrays



"""
The configuration system is used to change NengoDL's default behaviour in
various ways.

See `the documentation <https://www.nengo.ai/nengo-dl/config.html>`__ for more details.
"""

from nengo import Connection, Ensemble, Network, Probe, ensemble
from nengo.builder import Model
from nengo.exceptions import ConfigError, NetworkContextError
from nengo.params import BoolParam, Parameter


def configure_settings(**kwargs):
    """
    Pass settings to ``nengo_dl`` by setting them as parameters on the
    top-level Network config.

    The settings are passed as keyword arguments to ``configure_settings``;
    e.g., to set ``trainable`` use ``configure_settings(trainable=True)``.

    Parameters
    ----------
    trainable : bool or None
        Adds a parameter to Nengo Ensembles/Connections that controls
        whether or not they will be optimized by `.Simulator.fit`.
        Passing ``None`` will use the default ``nengo_dl`` trainable settings,
        or True/False will override the default for all objects.  In either
        case trainability can be further configured on a per-object basis (e.g.
        ``net.config[my_ensemble].trainable = True``.  See `the documentation
        <https://www.nengo.ai/nengo-dl/simulator.html#choosing-which-elements-to-optimize>`__
        for more details.
    planner : graph planning algorithm
        Pass one of the `graph planners
        <https://www.nengo.ai/nengo-dl/reference.html#graph-optimization>`_ to change
        the default planner.
    sorter : signal sorting algorithm
        Pass one of the `sort algorithms
        <https://www.nengo.ai/nengo-dl/reference.html#graph-optimization>`_ to change
        the default sorter.
    simplifications: list of graph simplification functions
        Pass a list of `graph simplification functions
        <https://www.nengo.ai/nengo-dl/reference.html#graph-optimization>`_ to change
        the default simplifications applied.  The default list of simplifications
        can be found in ``nengo_dl.graph_optimizer.default_simplifications``.
    inference_only : bool
        Set to True if the network will only be run in inference mode (i.e.,
        no calls to `.Simulator.fit`).  This may result in a small
        increase in the inference speed.
    lif_smoothing : float
        If specified, use the smoothed `.SoftLIFRate` neuron
        model, with the given smoothing parameter (``sigma``),
        to compute the gradient for `~nengo.LIF` neurons (as
        opposed to using `~nengo.LIFRate`).
    dtype : ``tf.DType``
        Set the floating point precision for simulation values.
    keep_history : bool
        Adds a parameter to Nengo Probes that controls whether or not they
        will keep the history from all simulation timesteps or only the last
        simulation step.  This can be further configured on a per-probe basis
        (e.g., ``net.config[my_probe].keep_history = False``).
    stateful : bool
        If True (default), the Simulator will be built to support stateful execution
        (where internal simulation state is preserved between simulator functions such
        as `.Simulator.predict`).  Otherwise all operations will be stateless. Note that
        this can also be configured individually through the ``stateful`` parameter on
        individual functions.
    use_loop : bool
        If True (default), use a symbolic while loop to run the simulation. Otherwise,
        simulation iterations are explicitly built into the model, avoiding the
        while loop. This can improve performance, but the simulation can only run
        for exactly ``Simulator.unroll_simulation`` iterations.
    """

    # get the toplevel network
    if len(Network.context) > 0:
        config = Network.context[0].config
    else:
        raise NetworkContextError(
            "`configure_settings` must be called within a Network context "
            "(`with nengo.Network(): ...`)"
        )

    try:
        params = config[Network]
    except ConfigError:
        config.configures(Network)
        params = config[Network]

    for attr, val in kwargs.items():
        if attr == "trainable":
            # for trainable, we set it on the nearest containing network (rather than
            # the top-level)
            sub_config = Network.context[-1].config
            for obj in (Ensemble, Connection, ensemble.Neurons):
                try:
                    obj_params = sub_config[obj]
                except ConfigError:
                    sub_config.configures(obj)
                    obj_params = sub_config[obj]

                obj_params.set_param(
                    "trainable", BoolParam("trainable", val, optional=True)
                )
        elif attr == "keep_history":
            config[Probe].set_param("keep_history", BoolParam("keep_history", val))
        elif attr in (
            "planner",
            "sorter",
            "simplifications",
            "inference_only",
            "lif_smoothing",
            "dtype",
            "stateful",
            "use_loop",
        ):
            params.set_param(attr, Parameter(attr, val))
        else:
            raise ConfigError("%s is not a valid config parameter" % attr)


def get_setting(model, setting, default=None, obj=None):
    """
    Returns config settings (created by `.configure_settings`).

    Parameters
    ----------
    model : `~nengo.builder.Model` or `~nengo.Network`
        Built model or Network containing all the config settings.
    setting : str
        Name of the config option to return.
    default
        The default value to return if config option not set.
    obj : ``NengoObject``
        The object on which config setting is stored (defaults to the top-level
        network).

    Returns
    -------
    config_val
        Value of ``setting`` if it has been specified, else ``default``.
    """

    if isinstance(model, Model):
        if model.toplevel is None:
            return default
        model = model.toplevel

    if obj is None:
        obj = model

    try:
        return getattr(model.config[obj], setting, default)
    except ConfigError:
        return default



"""
Utility objects used throughout the code base.
"""

from distutils.version import LooseVersion
import logging
import re
import subprocess
import sys
import threading
import time

from nengo.exceptions import SimulationError
import numpy as np
import progressbar
import tensorflow as tf


logger = logging.getLogger(__name__)

# check if GPU support is available
# note: we run this in a subprocess because list_physical_devices()
# will fix certain process-level TensorFlow configuration
# options the first time it is called
tf_gpu_installed = not subprocess.call(
    [
        sys.executable,
        "-c",
        "import sys; "
        "import tensorflow as tf; "
        "sys.exit(len(tf.config%s.list_physical_devices('GPU')) == 0)"
        % (".experimental" if LooseVersion(tf.__version__) < "2.1.0" else ""),
    ]
)


def sanitize_name(name):
    """
    Remove illegal TensorFlow name characters from string.

    Valid TensorFlow name characters are ``[A-Za-z0-9_.\\-/]``

    Parameters
    ----------
    name : str
        Name to be sanitized

    Returns
    -------
    sanitized : str
        Sanitized name
    """

    if not isinstance(name, str):
        name = str(name)

    name = name.replace(" ", "_")
    name = name.replace(":", "_")

    valid_exp = re.compile(r"[A-Za-z0-9_.\-/]")

    return "".join([c for c in name if valid_exp.match(c)])


def function_name(func, sanitize=True):
    """
    Get the name of the callable object ``func``.

    Parameters
    ----------
    func : callable
        Callable object (e.g., function, callable class)
    sanitize : bool
        If True, remove any illegal TensorFlow name characters from name

    Returns
    -------
    name : str
        Name of ``func`` (optionally sanitized)
    """

    name = getattr(func, "__name__", func.__class__.__name__)
    if sanitize:
        name = sanitize_name(name)

    return name


def align_func(output_shape, output_dtype):
    """
    Decorator that ensures the output of ``func`` is an
    `~numpy.ndarray` with the given shape and dtype.

    Parameters
    ----------
    output_shape : (list of) tuple of int
        Desired shape for function output(s) (must have the same size as actual
        function output)
    output_dtype : (list of) ``tf.DType`` or `~numpy.dtype`
        Desired dtype of function output(s)

    Raises
    ------
    ``nengo.exceptions.SimulationError``
        If the function returns ``None`` or a non-finite value.
    """

    single_output = isinstance(output_shape, tuple)

    if single_output:
        output_shape = [output_shape]
        output_dtype = [output_dtype]

    for i, dtype in enumerate(output_dtype):
        if isinstance(dtype, tf.DType):
            output_dtype[i] = dtype.as_numpy_dtype

    def apply_align(func):
        def aligned_func(*args):
            output = func(*args)

            if output is None:
                raise SimulationError(
                    "Function %r returned None" % function_name(func, sanitize=False)
                )

            if single_output:
                output = [output]

            for i, o in enumerate(output):
                try:
                    if not np.all(np.isfinite(o)):
                        raise SimulationError(
                            "Function %r returned invalid value %r"
                            % (function_name(func, sanitize=False), o)
                        )
                except (TypeError, ValueError):
                    raise SimulationError(
                        "Function %r returned a value %r of invalid type %r"
                        % (function_name(func, sanitize=False), o, type(o))
                    )
                o = np.asarray(o, dtype=output_dtype[i])
                o = o.reshape(output_shape[i])
                output[i] = o

            if single_output:
                output = output[0]

            return output

        return aligned_func

    return apply_align


def print_op(input, message):
    """
    Inserts a print statement into the TensorFlow graph.

    Parameters
    ----------
    input : ``tf.Tensor``
        The value of this tensor will be printed whenever it is computed
        in the graph
    message : str
        String prepended to the value of ``input``, to help with logging

    Returns
    -------
    op : ``tf.Tensor``
        New tensor representing the print operation applied to ``input``

    Notes
    -----
    This is what ``tf.Print`` is supposed to do, but it doesn't seem to work
    consistently.
    """

    def print_func(x):  # pragma: no cover (runs in TF)
        print(message, str(x))
        return x

    with tf.device("/cpu:0"):
        output = tf.numpy_function(print_func, [input], input.dtype)
    output.set_shape(input.shape)

    return output


class MessageBar(progressbar.BouncingBar):
    """
    ProgressBar widget for progress bars with possibly unknown duration.

    Parameters
    ----------
    msg : str
        A message to be displayed in the middle of the progress bar
    finish_msg : str
        A message to be displayed when the progress bar is finished
    """

    def __init__(self, msg="", finish_msg="", **kwargs):
        super().__init__(**kwargs)
        self.msg = msg
        self.finish_msg = finish_msg

    def __call__(self, progress, data, width):
        if progress.end_time:
            return self.finish_msg

        if progress.max_value is progressbar.UnknownLength:
            bar = progressbar.BouncingBar
        else:
            bar = progressbar.Bar
        line = bar.__call__(self, progress, data, width)

        if data["percentage"] is None:
            msg = self.msg
        else:
            msg = "%s (%d%%)" % (self.msg, data["percentage"])

        offset = width // 2 - len(msg) // 2

        return line[:offset] + msg + line[offset + len(msg) :]


class ProgressBar(progressbar.ProgressBar):  # pylint: disable=too-many-ancestors
    """
    Handles progress bar display for some tracked process.

    Parameters
    ----------
    present : str
        Description of process in present (e.g., "Simulating")
    past : str
        Description of process in past (e.g., "Simulation")
    max_value : int or None
        The maximum number of steps in the tracked process (or ``None`` if
        the maximum number of steps is unknown)

    Notes
    -----
    Launches a separate thread to handle the progress bar display updates.
    """

    def __init__(self, present="", past=None, max_value=1, **kwargs):

        self.present = present
        self.sub_bar = None
        self.finished = None

        if past is None:
            past = present

        self.msg_bar = MessageBar(msg=present, finish_msg="%s finished in" % past)
        widgets = [self.msg_bar, " "]

        if max_value is None:
            widgets.append(progressbar.Timer(format="%(elapsed)s"))
        else:
            widgets.append(
                progressbar.ETA(format="ETA: %(eta)s", format_finished="%(elapsed)s")
            )

        def update_thread():
            while not self.finished:
                if self.sub_bar is None or self.sub_bar.finished:
                    self.update()
                time.sleep(0.001)

        self.thread = threading.Thread(target=update_thread)
        self.thread.daemon = True

        if max_value is None:
            max_value = progressbar.UnknownLength

        super().__init__(
            poll_interval=0.1,
            widgets=widgets,
            fd=sys.stdout,
            max_value=max_value,
            **kwargs,
        )

    def start(self, **kwargs):
        """Start tracking process, initialize display."""

        super().start(**kwargs)

        self.finished = False
        self.thread.start()

        return self

    def finish(self, **kwargs):
        """Stop tracking process, finish display."""

        if self.sub_bar is not None and self.sub_bar.finished is False:
            self.sub_bar.finish()

        self.finished = True
        self.thread.join()

        super().finish(**kwargs)

    def step(self):
        """
        Advance the progress bar one step.
        """

        self.value += 1

    def sub(self, msg=None, **kwargs):
        """
        Creates a new progress bar for tracking a sub-process.

        Parameters
        ----------
        msg : str
            Description of sub-process
        """

        if self.sub_bar is not None and self.sub_bar.finished is False:
            self.sub_bar.finish()

        self.sub_bar = SubProgressBar(
            present="%s: %s" % (self.present, msg) if msg else self.present, **kwargs
        )

        return self.sub_bar

    @property
    def max_steps(self):
        """
        Alias for max_value to allow this to work with Nengo progress bar
        interface.
        """
        return self.max_value

    @max_steps.setter
    def max_steps(self, n):
        self.max_value = n

    def __enter__(self):
        super().__enter__()

        return self.start()

    def __next__(self):
        """Wraps an iterable using this progress bar."""

        try:
            assert self.start_time is not None
            self.step()
            value = next(self._iterable)
            return value
        except StopIteration:
            self.finish()
            raise


class SubProgressBar(ProgressBar):  # pylint: disable=too-many-ancestors
    """
    A progress bar representing a sub-task within an overall progress bar.
    """

    def finish(self, **kwargs):
        """Finishing a sub-progress bar doesn't start a new line."""
        super().finish(end="\r", **kwargs)


class NullProgressBar(progressbar.NullBar):  # pylint: disable=too-many-ancestors
    """
    A progress bar that does nothing.

    Used to replace ProgressBar when we want to disable output.
    """

    def __init__(self, present="", past=None, max_value=1, **kwargs):
        super().__init__(max_value=max_value, **kwargs)

    def sub(self, *args, **kwargs):
        """
        Noop for creating a sub-progress bar.
        """
        return self

    def step(self, **kwargs):
        """
        Noop for incrementing the progress bar.
        """



# pylint: disable=missing-docstring

from importlib import reload
import sys

from nengo import version as nengo_version
import pytest

from nengo_dl import version


def test_nengo_version_check():
    # note: we rely on travis-ci to test this against different nengo versions

    if version.dev is not None or nengo_version.dev is None:
        # nengo_dl should be compatible with all non-development nengo
        # versions, and a nengo_dl dev version should be compatible with all
        # (dev or non-dev) nengo versions
        assert nengo_version.version_info <= version.latest_nengo_version
        with pytest.warns(None) as w:
            reload(version)
        assert len(w) == 0
    else:
        # a development version of nengo with a non-development nengo_dl
        # version should cause a warning (we don't want to mark a nengo_dl
        # release as compatible with a nengo dev version, since the nengo
        # version may change and no longer be compatible with our nengo_dl
        # release).

        # note: we assume that a nengo dev version means the latest dev
        # version
        assert nengo_version.version_info >= version.latest_nengo_version

        with pytest.warns(UserWarning, match="This version.*not been tested"):
            reload(version)

    # check that it still works without nengo (faking an import error by
    # messing up sys.modules)
    saved = sys.modules["nengo.version"]
    sys.modules["nengo.version"] = None
    with pytest.warns(None) as w:
        reload(version)
    assert len(w) == 0
    sys.modules["nengo.version"] = saved



"""
The builder manages the mapping between (groups of) Nengo operators and the
builder objects that know how to translate those operators into a
TensorFlow graph.
"""

from collections import namedtuple
import logging
import warnings

from nengo import builder
from nengo.builder import signal
from nengo.exceptions import BuildError
import numpy as np
import tensorflow as tf

from nengo_dl import utils

logger = logging.getLogger(__name__)


class Builder:
    """
    Manages the operator build classes known to the ``nengo_dl`` build process.

    Parameters
    ----------
    plan : list of tuple of `~nengo.builder.Operator`
        The groups of operators that will be built
    signals : `.signals.SignalDict`
        Mapping from `~nengo.builder.Signal` to
        ``tf.Tensor`` (updated by operations)
    config : `.BuildConfig`
        Configuration parameters for the build process
    """

    builders = {}

    def __init__(self, plan, signals, config):
        self.plan = plan
        self.signals = signals
        self.config = config
        self.op_builds = {}

    def build_pre(self, progress=None):
        """
        Setup step for build classes, in which they compute any of the
        values that are constant across simulation timesteps.

        Parameters
        ----------
        progress : `.utils.ProgressBar`
            Progress bar for ops in plan
        """

        for ops in self.plan:
            logger.debug("===================")
            logger.debug("PRE BUILD %s", ops)
            logger.debug("sets %s", [op.sets for op in ops])
            logger.debug("incs %s", [op.incs for op in ops])
            logger.debug("reads %s", [op.reads for op in ops])
            logger.debug("updates %s", [op.updates for op in ops])

            if type(ops[0]) not in Builder.builders:
                raise BuildError(
                    "No registered builder for operators of type %r" % type(ops[0])
                )

            BuildClass = Builder.builders[type(ops[0])]

            with self.name_scope(ops):
                self.op_builds[ops] = BuildClass(ops, self.signals, self.config)

            if progress is not None:
                progress.step()

    def build(self, progress=None):
        """
        Build the computations implementing a single simulator timestep.

        Parameters
        ----------
        progress : `.utils.ProgressBar`
            Progress bar for ops in plan

        Returns
        -------
        side_effects : list of ``tf.Tensor``
            Outputs with possible side-effects, i.e. computations that need to
            be executed in the TensorFlow graph even if their output doesn't
            appear to be used.
        """

        side_effects = []
        for ops in self.plan:
            logger.debug("===================")
            logger.debug("BUILD %s", ops)

            with self.name_scope(ops):
                output = self.op_builds[ops].build_step(self.signals)

            if isinstance(output, (tf.Tensor, tf.Variable)):
                side_effects.append(output)
            elif isinstance(output, (list, tuple)):
                side_effects.extend(list(output))

            if progress is not None:
                progress.step()

        return side_effects

    def build_post(self, progress=None):
        """
        Calls post build functions for all ops in plan.

        Parameters
        ----------
        progress : `.utils.ProgressBar`
            Progress bar for ops in plan
        """

        for ops in self.plan:
            logger.debug("===================")
            logger.debug("POST BUILD %s", ops)

            with self.name_scope(ops):
                self.op_builds[ops].build_post(ops, self.signals, self.config)

            if progress is not None:
                progress.step()

    def name_scope(self, ops):
        """Returns a new TensorFlow name scope for the given ops."""

        return tf.name_scope(
            utils.sanitize_name(Builder.builders[type(ops[0])].__name__)
        )

    @classmethod
    def register(cls, nengo_op):
        """
        A decorator for adding a class to the build function registry.

        Parameters
        ----------
        nengo_op : `~nengo.builder.Operator`
            The operator associated with the build function being decorated.
        """

        def register_builder(build_class):
            if not issubclass(build_class, OpBuilder):
                warnings.warn("Build classes should inherit from OpBuilder")

            if nengo_op in cls.builders:
                warnings.warn(
                    "Operator '%s' already has a builder. Overwriting." % nengo_op
                )

            cls.builders[nengo_op] = build_class
            return build_class

        return register_builder


class BuildConfig(
    namedtuple(
        "BuildConfig",
        ("inference_only", "lif_smoothing", "cpu_only", "rng", "training"),
    )
):
    """
    Stores configuration parameters that may be relevant to parts of the
    build process.

    Parameters
    ----------
    inference_only : bool
        If True the network should be constructed in "inference only" mode
        (omitting any support for training operations).
    lif_smoothing : float
        Smoothing parameter for `~nengo.LIF` gradient approximation.
    cpu_only : bool
        True if TensorFlow is only running on the CPU (because that was
        specified by the user or because ``tensorflow-gpu`` is not installed).
    rng : `~numpy.random.mtrand.RandomState`
        Seeded random number generator.
    training : ``tf.Tensor`` (bool)
        True if building in training mode, False for inference mode.
    """

    __slots__ = ()


class OpBuilder:
    """
    Base class for build classes, which implement the logic for building a group of
    Nengo Operators into TensorFlow.

    The constructor should set up any computations that are fixed for
    this op (i.e., things that do not need to be recomputed each timestep).

    Parameters
    ----------
    ops : list of `~nengo.builder.Operator`
        The operator group to build into the model
    signals : `.signals.SignalDict`
        Mapping from `~nengo.builder.Signal` to
        ``tf.Tensor`` (updated by operations)
    config : `~.builder.BuildConfig`
        General repository for config information builders might want
        (conglomerated into this object so that we can add/remove config data
        without having to change the function signature all the time).
    """

    def __init__(self, ops, signals, config):
        logger.debug(self.__class__.__name__)
        logger.debug("\n".join(str(x) for x in ops))

        self.config = config

    def build_step(self, signals):
        """
        This function builds whatever computations need to be executed in
        each simulation timestep.

        Parameters
        ----------
        signals : `.signals.SignalDict`
            Mapping from `~nengo.builder.Signal` to
            ``tf.Tensor`` (updated by operations)

        Returns
        -------
        side_effects : list of ``tf.Tensor``
            If not None, the returned tensors correspond to outputs with
            possible side-effects, i.e. computations that need to be executed
            in the TensorFlow graph even if their output doesn't appear to be
            used
        """
        raise BuildError("OpBuilders must implement a `build_step` function")

    def build_post(self, ops, signals, config):
        """
        This function will be called after the graph has been built and
        each time the Simulator is reset.

        Note that this function may be called multiple times per session, so
        it should do any required operations in-place.

        Parameters
        ----------
        ops : list of `~nengo.builder.Operator`
            The operator group to build into the model
        signals : `.signals.SignalDict`
            Mapping from `~nengo.builder.Signal` to
            ``tf.Tensor`` (updated by operations)
        config : `~.builder.BuildConfig`
            General repository for config information builders might want
            (conglomerated into this object so that we can add/remove config data
            without having to change the function signature all the time).
        """

    @staticmethod
    def mergeable(x, y):
        """
        Compute the mergeability of two operators of this builder's type.

        Parameters
        ----------
        x : `nengo.builder.Operator`
            The operator being tested
        y : `nengo.builder.Operator`
            The operator being merged into (this is representative of a group
            of operators that have already been merged)

        Returns
        -------
        mergeable : bool
            True if ``x`` and ``y`` can be merged into a single built op,
            else ``False``.
        """

        return False


class NengoBuilder(builder.Builder):
    """
    Copy of the default Nengo builder.

    This class is here so that we can register new build functions for
    Nengo DL without affecting the default Nengo build process.
    """

    builders = {}

    @classmethod
    def build(cls, model, obj, *args, **kwargs):
        """
        Build ``obj`` into ``model``.

        This method looks up the appropriate build function for ``obj`` and
        calls it with the model and other arguments provided.

        In addition to the parameters listed below, further positional and
        keyword arguments will be passed unchanged into the build function.

        Parameters
        ----------
        model : Model
            The `~nengo.builder.Model` instance in which to store build
            artifacts.
        obj : object
            The object to build into the model.
        """

        try:
            # first try building obj using any custom build functions that have
            # been registered by Nengo DL
            return builder.Builder.build.__func__(
                NengoBuilder, model, obj, *args, **kwargs
            )
        except BuildError:
            # fallback on normal nengo builder
            return builder.Builder.build.__func__(
                builder.Builder, model, obj, *args, **kwargs
            )


class NengoModel(builder.Model):
    """
    Copy of the default Nengo model.

    This allows us to override certain model behaviours.

    Parameters
    ----------
    fail_fast : bool
        If True, try to call ``op.make_step`` when ops are added to the model.
        Note that NengoDL doesn't actually use ``make_step``, so errors in that
        function are not necessarily errors in NengoDL (which is why we want to
        disable that check). But it might still be useful when debugging
        new op/build functions, which is why we leave the option.
    """

    def __init__(self, *args, fail_fast=True, **kwargs):
        self.fail_fast = fail_fast
        super().__init__(*args, **kwargs)

    def add_op(self, op):
        """
        Add an operator to the model.

        Parameters
        ----------
        op : `~nengo.builder.Operator`
            Operator being added to the model.

        Notes
        -----
        This is a copy of the parent `nengo.builder.Model.add_op`, with the
        addition of the ``if self.fail_fast`` condition.
        """

        # TODO: nengo 3.0 adds something similar to this condition, but
        # it uses an rc setting (so we can't change it in nengo-dl without
        # also changing nengo core). if the rc system is reworked to allow
        # backend-specific overrides, we could remove this class.

        self.operators.append(op)
        if self.fail_fast:
            # Fail fast by trying make_step with a temporary sigdict
            signals = signal.SignalDict()
            op.init_signals(signals)
            op.make_step(signals, self.dt, np.random)



"""
The Simulator class is the access point for the main features of NengoDL,
including `running <.Simulator.run_steps>` and `training <.Simulator.fit>`
a model.
"""

import collections
import copy
from functools import partial
import logging
import textwrap
import warnings

import jinja2
from nengo import Connection, Direct, Ensemble, Network, Node, Probe, Convolution
from nengo import rc as nengo_rc
from nengo.builder.connection import BuiltConnection
from nengo.builder.ensemble import BuiltEnsemble
from nengo.ensemble import Neurons
from nengo.exceptions import (
    ReadonlyError,
    SimulatorClosed,
    NengoWarning,
    SimulationError,
    ValidationError,
)
from nengo.solvers import NoSolver
from nengo.utils.magic import decorator
import numpy as np
import tensorflow as tf

from nengo_dl import callbacks, compat, config, utils
from nengo_dl.builder import NengoBuilder, NengoModel
from nengo_dl.tensor_graph import TensorGraph


logger = logging.getLogger(__name__)


@decorator
def with_self(wrapped, instance, args, kwargs):
    """A decorator that can be used to ensure that any TensorFlow operations happening
    within a method will use the settings associated with this Simulator."""

    keras_dtype = tf.keras.backend.floatx()
    tf.keras.backend.set_floatx(instance.tensor_graph.dtype)
    with instance.graph.as_default(), instance.graph.device(
        instance.tensor_graph.device
    ):
        output = wrapped(*args, **kwargs)
    tf.keras.backend.set_floatx(keras_dtype)

    return output


@decorator
def require_open(wrapped, instance, args, kwargs):
    """A decorator that can be used to mark methods that require the Simulator to
    be open."""

    if instance.closed:
        raise SimulatorClosed(
            "Cannot call %s after simulator is closed" % wrapped.__name__
        )

    return wrapped(*args, **kwargs)


def fill_docs(*args, **kwargs):
    """Stores documentation for common arguments in one place, to avoid duplication,
    and then fills them in automatically in the docstring."""

    docs = {
        "x": """
        {% set uses_y = func_name in ("fit", "evaluate") %}

        {% if func_name in ("predict_on_batch", "run_steps") %}
            {% set batch_size = 1 %}
        {% else %}
            {% set batch_size = 50 %}
        {% endif %}

        Data for input Nodes in the model. This argument is optional; if
        it is not specified, then data will automatically be generated
        according to the inputs specified in the Node definitions (e.g., by calling
        the output function associated with that Node).

        ``{{ param_name }}`` can be specified as:

        - A dictionary of {`nengo.Node` or str: `numpy.ndarray`}
          indicating the input values for the given nodes. Nodes can be referred
          to by the Node object itself or by a string name, which will be
          ``Node.label`` if one was specified or ``"node"``
          (duplicate names will have a number appended, corresponding to the order
          found in `nengo.Network.all_nodes`).
        - A list of `numpy.ndarray` indicating the input values for each
          input Node, ordered according to the order in which the Nodes were
          added to the model (this corresponds to the order found in
          `nengo.Network.all_nodes`).
        - A `numpy.ndarray` indicating the input value for a single input Node.
        {% if func_name != "run_steps" %}
        - A generator or ``tf.data.Dataset`` that produces one of the above.
        {% endif %}

        All inputs should have shape ``(batch_size, n_steps, node.size_out)``.

        For example, if the model only has a single input Node, then
        ``{{ param_name }}`` can simply be an ndarray of data for that Node.

        .. testcode::

            with nengo.Network() as net:
                a = nengo.Node([0])
                p = nengo.Probe(a)

            with nengo_dl.Simulator(net) as sim:
                {% if uses_y %}
                sim.compile(loss="mse")
                sim.{{ func_name }}(
                    {{ param_name }}=np.ones((50, 10, 1)), y=np.ones((50, 10, 1)))
                {% elif func_name == "run_steps" %}
                sim.{{ func_name }}(
                    10, {{ param_name }}=np.ones(({{ batch_size }}, 10, 1)))
                {% else %}
                sim.{{ func_name }}(
                    {{ param_name }}=np.ones(({{ batch_size }}, 10, 1)))
                {% endif %}

        {% if uses_y %}
        .. testoutput::
            :hide:

            ...
        {% endif %}

        If the network has multiple inputs, then ``{{ param_name }}`` can be specified
        as a dictionary mapping `nengo.Node` objects to arrays, e.g.

        .. testcode::

            with nengo.Network() as net:
                a = nengo.Node([0])
                b = nengo.Node([0, 0])
                p = nengo.Probe(a)

            with nengo_dl.Simulator(net) as sim:
                {% if uses_y %}
                sim.compile(loss="mse")
                sim.{{ func_name }}(
                    {{ param_name }}={
                        a: np.ones((50, 10, 1)),
                        b: np.ones((50, 10, 2))
                    },
                    y=np.ones((50, 10, 1))
                )
                {% elif func_name == "run_steps" %}
                sim.{{ func_name }}(
                    10,
                    {{ param_name }}={
                        a: np.ones(({{ batch_size }}, 10, 1)),
                        b: np.ones(({{ batch_size }}, 10, 2))
                    }
                )
                {% else %}
                sim.{{ func_name }}(
                    {{ param_name }}={
                        a: np.ones(({{ batch_size }}, 10, 1)),
                        b: np.ones(({{ batch_size }}, 10, 2))
                    }
                )
                {% endif %}

        {% if uses_y %}
        .. testoutput::
            :hide:

            ...
        {% endif %}

        If an input value is not specified for one of the Nodes in the model then
        data will be filled in automatically according to the Node definition.

        {% if func_name != "run_steps" %}
        For dynamic input types (e.g., ``tf.data`` pipelines or generators), NengoDL
        tries to avoid introspecting/altering the data before the
        simulation starts, as this may have unintended side-effects. So data must be
        specified via one of the standard Keras methods (arrays, list of arrays, or
        string name dictionary; using a dictionary of Node objects is not supported).
        In addition, data must be explicitly provided for all input nodes (it will not
        be automatically generated if data is not specified).

        In addition, when using dynamic inputs, data must be provided for the
        special ``"n_steps"`` input. This specifies the number of timesteps that the
        simulation will run for. Technically this is just a single scalar value
        (e.g., ``10``). But Keras requires that all input data be batched, so that
        input value needs to be duplicated into an array with size
        ``(batch_size, 1)`` (where all entries have the same value, e.g. ``10``).

        {% if uses_y %}
        Also keep in mind that when using a dynamic input for ``x`` the ``y`` parameter
        is unused, and instead the generator should return ``(x, y)`` pairs.
        {% endif %}

        .. testcode::

            with nengo.Network() as net:
                a = nengo.Node([0], label="a")
                p = nengo.Probe(a, label="p")

            with nengo_dl.Simulator(net) as sim:
                {% if uses_y %}
                dataset = tf.data.Dataset.from_tensor_slices(
                    ({"a": tf.ones((50, 10, 1)),
                      "n_steps": tf.ones((50, 1), dtype=tf.int32) * 10},
                     {"p": tf.ones((50, 10, 1))})
                ).batch(sim.minibatch_size)

                sim.compile(loss="mse")
                sim.{{ func_name }}({{ param_name }}=dataset)
                {% else %}
                dataset = tf.data.Dataset.from_tensor_slices(
                    {"a": tf.ones((50, 10, 1)),
                     "n_steps": tf.ones((50, 1), dtype=tf.int32) * 10}
                ).batch(sim.minibatch_size)

                sim.{{ func_name }}({{ param_name }}=dataset)
                {% endif %}

        {% if uses_y %}
        .. testoutput::
            :hide:

            ...
        {% endif %}
        {% endif %}
        """,
        "y": """
        Target values for Probes in the model. These can be specified in the same
        ways as the input values in ``x``, except using Probes instead of Nodes.
        All targets should have shape ``(batch_size, n_steps, probe.size_in)``.

        For example,

        .. testcode::

            with nengo.Network() as net:
                a = nengo.Node([0])
                p = nengo.Probe(a)

            with nengo_dl.Simulator(net) as sim:
                sim.compile(loss="mse")
                sim.{{ func_name }}(
                    x={a: np.zeros((50, 10, 1))}, y={p: np.zeros((50, 10, 1))})

        .. testoutput::
            :hide:

            ...

        Note that data is only specified for the probes used in the loss function
        (specified when calling `.Simulator.compile`).  For example, if we have two
        probes, but only one is used during training (the other is used for data
        collection during inference), we could set this up like:

        .. testcode::

            with nengo.Network() as net:
                a = nengo.Node([0])
                b = nengo.Node([0])
                p_a = nengo.Probe(a)
                p_b = nengo.Probe(b)

            with nengo_dl.Simulator(net) as sim:
                # compiled loss function only depends on p_a
                sim.compile(loss={p_a: "mse"})

                # only specify data for p_a
                sim.{{ func_name }}(
                    x={a: np.zeros((50, 10, 1))},  y={p_a: np.zeros((50, 10, 1))})

        .. testoutput::
            :hide:

            ...

        ``y`` is not used if ``x`` is a generator. Instead, the generator passed to
        ``x`` should yield ``(x, y)`` tuples, where ``y`` is in one of the formats
        described above.
        """,
        "n_steps": """
        The number of simulation steps to be executed.  This parameter is optional;
        if not specified, the number of simulation steps will be inferred from the
        input data. However, this parameter can be useful if you don't want to
        specify input data (you just want to use the inputs defined by the Nengo
        Nodes), or if your model does not have any input Nodes (so there is no data
        to be passed in).
        """,
        "stateful": """
        This parameter controls whether or not the saved internal stimulation state
        will be updated after a run completes. If ``stateful=False`` then the initial
        state of future runs will be unaffected by this run. With ``stateful=True``,
        future runs will begin from the terminal state of this run.

        For example,

        .. code-block:: python

            # begins in state0, terminates in state1
            sim.{{ func_name }}(..., stateful=False)
            # begins in state0, terminates in state2
            sim.{{ func_name }}(..., stateful=True)
            # begins in state2, terminates in state3
            sim.{{ func_name }}(..., stateful=False)
            # begins in state2, terminates in state4
            sim.{{ func_name }}(..., stateful=True)

        Note that `.Simulator.reset` can be used to reset the state to initial
        conditions at any point.
        """,
    }

    # use default name for args
    for arg in args:
        kwargs[arg] = arg

    env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)

    def fill_documentation(func):
        rendered_docs = {}
        for name, template in kwargs.items():
            doc = docs[template]

            # fill in variables
            doc = env.from_string(doc).render(param_name=name, func_name=func.__name__)

            # correct indentation
            doc = textwrap.indent(doc, " " * 4)
            doc = doc.strip()

            rendered_docs[name] = doc

        # insert result into docstring
        func.__doc__ = env.from_string(func.__doc__).render(**rendered_docs)

        return func

    return fill_documentation


class Simulator:  # pylint: disable=too-many-public-methods
    """
    Simulate network using the ``nengo_dl`` backend.

    Parameters
    ----------
    network : `nengo.Network`
        A network object to be built and then simulated.
    dt : float
        Length of a simulator timestep, in seconds.
    seed : int
        Seed for all stochastic operators used in this simulator.
    model : `~nengo.builder.Model`
        Pre-built model object (mainly used for debugging).
    device : None or ``"/cpu:0"`` or ``"/gpu:[0-n]"``
        This specifies the computational device on which the simulation will
        run.  The default is ``None``, which means that operations will be assigned
        according to TensorFlow's internal logic (generally speaking, this means that
        things will be assigned to the GPU if ``tensorflow-gpu`` is installed,
        otherwise everything will be assigned to the CPU).  The device can be set
        manually by passing the `TensorFlow device specification
        <https://www.tensorflow.org/api_docs/python/tf/Graph#device>`_ to this
        parameter.  For example, setting ``device="/cpu:0"`` will force everything
        to run on the CPU.  This may be worthwhile for small models, where the extra
        overhead of communicating with the GPU outweighs the actual computations.  On
        systems with multiple GPUs, ``device="/gpu:0"``/``"/gpu:1"``/etc. will select
        which one to use.
    unroll_simulation : int
        This controls how many simulation iterations are executed each time through
        the outer simulation loop.  That is, we could run 20 timesteps as

        .. code-block:: python

            for i in range(20):
                <run 1 step>

        or

        .. code-block:: python

            for i in range(5):
                <run 1 step>
                <run 1 step>
                <run 1 step>
                <run 1 step>

        This is an optimization process known as "loop unrolling", and
        ``unroll_simulation`` controls how many simulation steps are unrolled.  The
        first example above would correspond to ``unroll_simulation=1``, and the
        second would be ``unroll_simulation=4``.

        Unrolling the simulation will result in faster simulation speed, but increased
        build time and memory usage.

        In general, unrolling the simulation will have no impact on the output of a
        simulation.  The only case in which unrolling may have an impact is if
        the number of simulation steps is not evenly divisible by
        ``unroll_simulation``.  In that case extra simulation steps will be executed,
        which could change the internal state of the simulation and
        will affect any subsequent calls to ``sim.run``.  So it is recommended that the
        number of steps always be evenly divisible by ``unroll_simulation``.
    minibatch_size : int
        The number of simultaneous inputs that will be passed through the
        network. For example, a single call to `.Simulator.run` will process
        ``minibatch_size`` input instances in parallel. Or when calling
        `.Simulator.predict`/`.Simulator.fit` with a batch of data, that data will be
        divided up into ``minibatch_size`` chunks.
    progress_bar : bool
        If True (default), display progress information when building a model. This will
        also be the default for the ``progress_bar`` argument within `.Simulator.run`
        and `.Simulator.run_steps`.

    Attributes
    ----------
    data : `.SimulationData`
        Stores simulation data and parameter values (in particular, the recorded output
        from probes after calling `.Simulator.run` can be accessed through
        ``sim.data[my_probe]``).
    model : `nengo.builder.Model`
        Built Nengo model, containing the data that defines the network to be simulated.
    keras_model : ``tf.keras.Model``
        Keras Model underlying the simulation (implements the inference/training loops).
    tensor_graph : `.tensor_graph.TensorGraph`
        Keras Layer implementing the Nengo simulation (built into ``keras_model``).
    """

    def __init__(
        self,
        network,
        dt=0.001,
        seed=None,
        model=None,
        device=None,
        unroll_simulation=1,
        minibatch_size=None,
        progress_bar=True,
    ):
        self.closed = None
        self.unroll = unroll_simulation
        self.minibatch_size = 1 if minibatch_size is None else minibatch_size
        self.data = SimulationData(self, minibatch_size is not None)

        if seed is None:
            if network is not None and network.seed is not None:
                seed = network.seed + 1
            else:
                seed = np.random.randint(np.iinfo(np.int32).max)

        if device is None and not utils.tf_gpu_installed:
            warnings.warn(
                "No GPU support detected. It is recommended that you "
                "install tensorflow-gpu (`pip install tensorflow-gpu`)."
            )
            logger.info("Running on CPU")
        else:
            logger.info(
                "Running on %s",
                "CPU/GPU" if device is None else ("CPU" if "cpu" in device else "GPU"),
            )

        self.progress_bar = progress_bar
        ProgressBar = utils.ProgressBar if progress_bar else utils.NullProgressBar

        # build model (uses default nengo builder)
        nengo_precision = nengo_rc.get("precision", "bits")
        nengo_rc.set(
            "precision",
            "bits",
            config.get_setting(model or network, "dtype", "float32")[-2:],
        )
        if model is None:
            self.model = NengoModel(
                dt=float(dt),
                label="%s, dt=%f" % (network, dt),
                builder=NengoBuilder(),
                fail_fast=False,
            )
        else:
            if dt != model.dt:
                warnings.warn(
                    "Model dt (%g) does not match Simulator "
                    "dt (%g)" % (model.dt, dt),
                    NengoWarning,
                )
            self.model = model

        if network is not None:
            p = ProgressBar("Building network", "Build")
            self.model.build(network, progress=p)

        nengo_rc.set("precision", "bits", nengo_precision)

        self.stateful = config.get_setting(self.model, "stateful", True)

        # set up tensorflow graph plan
        with ProgressBar(
            "Optimizing graph", "Optimization", max_value=None
        ) as progress:
            self.tensor_graph = TensorGraph(
                self.model,
                self.dt,
                unroll_simulation,
                self.minibatch_size,
                device,
                progress,
                seed,
            )

        # build keras models
        with ProgressBar(
            "Constructing graph", "Construction", max_value=None
        ) as progress:
            self.graph = tf.Graph()
            self._build_keras(progress)

        # initialize sim attributes
        self._n_steps = self._time = 0
        for p in self.model.probes:
            self.model.params[p] = []

        self.closed = False

    @with_self
    def _build_keras(self, progress=None):
        """
        Build the underlying Keras model that drives the simulation.

        Parameters
        ----------
        progress : `.utils.ProgressBar`
            Progress bar for construction stage.
        """

        tf.config.set_soft_device_placement(False)

        self.node_inputs, n_steps = self.tensor_graph.build_inputs()
        inputs = list(self.node_inputs.values()) + [n_steps]

        outputs = self.tensor_graph(
            inputs,
            stateful=self.stateful,
            # if the global learning phase is set, use that
            training=compat.global_learning_phase(),
            progress=progress,
        )

        self.keras_model = tf.keras.Model(
            inputs=inputs, outputs=outputs, name="keras_model",
        )

        # set more informative output names
        # keras names them like LayerName_i, whereas we would like to have the names
        # associated with the probes
        self.keras_model.output_names = [
            self.tensor_graph.io_names[p] for p in self.model.probes
        ] + ["steps_run"]

        self.tensor_graph.build_post()

    @require_open
    @with_self
    def reset(
        self,
        seed=None,
        include_trainable=True,
        include_probes=True,
        include_processes=True,
    ):
        """
        Resets the simulator to initial conditions.

        Parameters
        ----------
        seed : int
            If not None, overwrite the default simulator seed with this value
            (note: this becomes the new default simulator seed).
        include_trainable : bool
            If True (default), also reset any online or offline training that has been
            performed on simulator parameters (e.g., connection weights).
        include_probes : bool
            If True (default), also clear probe data.
        include_processes: bool
            If True (default), also reset all `nengo.Process` objects in the model.

        Notes
        -----
        Changing the TensorFlow seed only affects ops created from then on; it has
        no impact on existing ops (either changing their seed or resetting their random
        state). So calling `.Simulator.reset` will likely have no impact on any
        TensorFlow randomness (it will still affect numpy randomness, such as in a
        `nengo.Process`, as normal).
        """

        if self.stateful:
            tf.keras.backend.batch_get_value(
                [var.initializer for var in self.tensor_graph.saved_state.values()]
            )

        if include_trainable:
            tf.keras.backend.batch_get_value(
                [var.initializer for var in self.tensor_graph.base_params.values()]
            )

        if include_probes:
            for p in self.model.probes:
                self.model.params[p] = []

        self._update_steps()

        # update rng
        if seed is not None:
            warnings.warn(
                "Changing the seed will not affect any TensorFlow operations "
                "created before the seed was updated"
            )
            self.tensor_graph.seed = seed

        if include_processes:
            self.tensor_graph.build_post()

    @require_open
    @with_self
    def soft_reset(self, include_trainable=False, include_probes=False):
        """
        Deprecated, use `.Simulator.reset` instead.
        """

        warnings.warn(
            "Simulator.soft_reset is deprecated, use Simulator.reset("
            "include_trainable=False, include_probes=False, include_processes=False) "
            "instead",
            DeprecationWarning,
        )
        self.reset(
            seed=None,
            include_trainable=include_trainable,
            include_probes=include_probes,
            include_processes=False,
        )

    @require_open
    @fill_docs("x", "n_steps", "stateful")
    def predict(self, x=None, n_steps=None, stateful=False, **kwargs):
        """
        Generate output predictions for the input samples.

        Computation is (optionally) done in batches.

        This function implements the `tf.keras.Model.predict
        <https://www.tensorflow.org/api_docs/python/tf/keras/Model#predict>`_ API.

        Parameters
        ----------
        x
            {{ x }}
        n_steps : int
            {{ n_steps }}
        stateful : bool
            {{ stateful}}
        kwargs: dict
            Will be passed on to `tf.keras.Model.predict
            <https://www.tensorflow.org/api_docs/python/tf/keras/Model#predict>`_.

        Returns
        -------
        probe_values : dict of {`nengo.Probe`: `numpy.ndarray`}
            Output values from all the Probes in the network.
        """

        return self._call_keras(
            "predict", x=x, n_steps=n_steps, stateful=stateful, **kwargs
        )

    @require_open
    @fill_docs("x", "n_steps", "stateful")
    def predict_on_batch(self, x=None, n_steps=None, stateful=False, **kwargs):
        """
        Generate output predictions for a single minibatch of input samples.

        Batch size is determined by ``sim.minibatch_size`` (i.e., inputs must have
        shape ``(sim.minibatch_size, n_steps, node.size_in)``.

        This function implements the `tf.keras.Model.predict_on_batch
        <https://www.tensorflow.org/api_docs/python/tf/keras/Model#predict_on_batch>`_
        API.

        Parameters
        ----------
        x
            {{ x }}
        n_steps : int
            {{ n_steps }}
        stateful : bool
            {{ stateful }}
        kwargs: dict
            Will be passed on to `tf.keras.Model.predict_on_batch
            <https://www.tensorflow.org/api_docs/python/tf/keras/Model#predict_on_batch>`_.

        Returns
        -------
        probe_values : dict of {`nengo.Probe`: `numpy.ndarray`}
            Output values from all the Probes in the network.
        """

        # need to reset if simulator is stateful but this call is not stateful
        need_reset = not stateful and self.stateful

        # predict_on_batch doesn't support callbacks, so we do it manually
        if need_reset:
            cbk = callbacks.IsolateState(self)

        # note: setting stateful to self.stateful so that the inner _call_keras won't
        # try to do any resetting
        output = self._call_keras(
            "predict_on_batch", x=x, n_steps=n_steps, stateful=self.stateful, **kwargs
        )

        if need_reset:
            cbk.reset()
            self._update_steps()

        return output

    @require_open
    @with_self
    def compile(self, *args, loss=None, metrics=None, loss_weights=None, **kwargs):
        """
        Configure the model for training/evaluation.

        Parameters
        ----------
        args
            Will be passed on to `tf.keras.Model.compile
            <https://www.tensorflow.org/api_docs/python/tf/keras/Model#compile>`_.
        loss
            Loss functions define the error that will be minimized during
            training.

            Losses can be specified as:

            - A `tf.losses.Loss
              <https://www.tensorflow.org/api_docs/python/tf/keras/losses>`_ instance.
            - A string matching the name of one of the loss functions above.
            - A function that accepts two arguments (``y_true, y_pred``) and returns
              a loss value (represented as a ``tf.Tensor``).
            - A list of some combination of the above, indicating different loss
              functions for each output Probe (ordered according to the order in
              which Probes were added to the model, which corresponds to the order
              found in ``Simulator.model.probes``).
            - A dictionary mapping Probe instances or names to loss functions.

            The total loss minimized during training will be the sum over the loss
            computed on each Probe (possibly weighted by ``loss_weights``).

            For example,

            .. testcode::

                with nengo.Network() as net:
                    node0 = nengo.Node([0])
                    node1 = nengo.Node([0])
                    probe0 = nengo.Probe(node0)
                    probe1 = nengo.Probe(node1)

                with nengo_dl.Simulator(net) as sim:
                    sim.compile(loss={probe0: "mse", probe1: tf.losses.mae})

            would compile ``probe0`` to use mean squared error and ``probe1`` to use
            mean absolute error.

        metrics
            Metrics are additional values (generally different kinds of losses) that
            will be computed during training for tracking purposes, but do not affect
            the result of the training.

            They can be specified in all the same ways as ``loss`` above.

            In addition, multiple metrics can be specified for each output Probe when
            using a list or dict, by providing multiple functions in a list (e.g.,
            ``metrics={my_probe: ["mae", "mse"]}``).
        loss_weights : list or dict
            Scalar weights that will be applied to the loss value computed for each
            output probe before summing them to compute the overall training loss. Can
            be a list (order corresponding to the order in ``loss``) or a dict mapping
            Probe instances/names to weights.
        kwargs
            Will be passed on to `tf.keras.Model.compile
            <https://www.tensorflow.org/api_docs/python/tf/keras/Model#compile>`_.
        """

        # convert inputs to canonical name dict form
        loss = self._standardize_data(loss, self.model.probes, broadcast_unary=True)
        metrics = self._standardize_data(
            metrics, self.model.probes, broadcast_unary=True
        )
        loss_weights = self._standardize_data(loss_weights, self.model.probes)

        self.keras_model.compile(
            *args, loss=loss, metrics=metrics, loss_weights=loss_weights, **kwargs
        )

    @require_open
    @fill_docs("x", "y", "n_steps", "stateful")
    def fit(self, x=None, y=None, n_steps=None, stateful=False, **kwargs):
        """
        Trains the model on some dataset.

        Note that if the model contains spiking neurons, during the execution of this
        function those neurons will be swapped for the equivalent non-spiking
        implementation (as opposed to, e.g., `Simulator.evaluate`, which will
        use the spiking implementation).

        Optimizer and loss functions are defined separately in `.Simulator.compile`.

        This function implements the `tf.keras.Model.fit
        <https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit>`_ API.

        Parameters
        ----------
        x
            {{ x }}
        y
            {{ y }}
        n_steps : int
            {{ n_steps }}
        stateful : bool
            {{ stateful }}
        kwargs: dict
            Will be passed on to `tf.keras.Model.fit
            <https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit>`_.

        Returns
        -------
        history : ``tf.keras.callbacks.History``
            The history has two attributes: ``history.epoch`` is the list of epoch
            numbers, and ``history.history`` is a dictionary keyed by metric names
            (e.g., "loss") containing a list of values of those metrics from each epoch.
        """

        # if validation data is None or a dataset we don't do anything, but
        # otherwise we apply the same data augmentation/validation
        # as for x and y
        if isinstance(kwargs.get("validation_data", None), (list, tuple)):
            validation_data = kwargs["validation_data"]

            x_val = validation_data[0]
            x_val = self._generate_inputs(x_val, n_steps=n_steps)
            self._check_data(x_val, n_steps=n_steps)

            y_val = validation_data[1]
            y_val = self._standardize_data(y_val, self.model.probes)
            self._check_data(y_val, n_steps=None, nodes=False)

            if len(validation_data) == 2:
                kwargs["validation_data"] = (x_val, y_val)
            else:
                kwargs["validation_data"] = (x_val, y_val, validation_data[2])

        return self._call_keras(
            "fit", x=x, y=y, n_steps=n_steps, stateful=stateful, **kwargs
        )

    @require_open
    @fill_docs("x", "y", "n_steps", "stateful")
    def evaluate(self, x=None, y=None, n_steps=None, stateful=False, **kwargs):
        """
        Compute the loss and metric values for the network.

        Loss functions and other metrics are defined separately in `.Simulator.compile`.

        This function implements the `tf.keras.Model.evaluate
        <https://www.tensorflow.org/api_docs/python/tf/keras/Model#evaluate>`_ API.

        Parameters
        ----------
        x
            {{ x }}
        y
            {{ y }}
        n_steps : int
            {{ n_steps }}
        stateful : bool
            {{ stateful }}
        kwargs: dict
            Will be passed on to `tf.keras.Model.evaluate
            <https://www.tensorflow.org/api_docs/python/tf/keras/Model#evaluate>`_.

        Returns
        -------
        outputs : dict of {str: `numpy.ndarray`}
            Computed loss/metric values. The overall loss will be in
            ``outputs["loss"]``, and values for each Probe will be in
            ``outputs["probe_name_loss"]`` or ``outputs["probe_name_metric_name"]``.
        """

        return self._call_keras(
            "evaluate", x=x, y=y, n_steps=n_steps, stateful=stateful, **kwargs
        )

    @with_self
    def _call_keras(
        self, func_type, x=None, y=None, n_steps=None, stateful=False, **kwargs
    ):
        """
        Internal base function for all the predict, fit, and evaluate functions.

        Parameters
        ----------
        func_type : "predict" or "predict_on_batch" or "fit" or "evaluate"
            The underlying function to call on the Keras model.
        x
            See description in documentation of ``<func_type>`` method.
        y
            See description in documentation of ``<func_type>`` method.
        n_steps : int
            See description in documentation of ``<func_type>`` method.
        stateful : bool
            See description in documentation of ``<func_type>`` method.
        kwargs : dict
            Will be passed to the underlying Keras function.

        Returns
        -------
            See description in documentation of ``<func_type>`` method.
        """

        if func_type.startswith("fit") and self.tensor_graph.inference_only:
            raise SimulationError(
                "Network was created with inference_only=True, cannot "
                "be run in training mode"
            )

        if stateful and not self.stateful:
            warnings.warn(
                "Ignoring stateful=True, since the model was created with the "
                "stateful=False config setting."
            )

        if "batch_size" in kwargs:
            # note: the keras "batch size" parameter refers to minibatch size
            # (i.e., the number of elements passed to the network in each iteration,
            # rather than the total number of elements in the data)
            warnings.warn(
                "Batch size is determined statically via Simulator.minibatch_size; "
                "ignoring value passed to `%s`" % func_type
            )
            kwargs["batch_size"] = None

        # TODO: apply standardize/generate/check data to generator somehow
        # maybe move it into a callback where the generated data is available?

        x = self._generate_inputs(x, n_steps=n_steps)
        self._check_data(
            x,
            n_steps=n_steps,
            batch_size=self.minibatch_size if "on_batch" in func_type else None,
        )

        if isinstance(x, dict):
            input_steps = x["n_steps"][0, 0]
            input_batch = x["n_steps"].shape[0]
        else:
            input_steps = None
            input_batch = self.minibatch_size if "on_batch" in func_type else None

        if y is not None:
            y = self._standardize_data(y, self.model.probes)
            # we set n_steps=None because targets do not necessarily need to have
            # the same number of timesteps as input (depending on the loss function)
            self._check_data(y, n_steps=None, batch_size=input_batch, nodes=False)

        if kwargs.get("validation_split", 0) != 0 and input_batch is not None:
            # validation_split is only a kwarg in `fit`, but we do it here because
            # we need to know `input_batch`.
            # split math set up to match
            # `keras.engine.training_utils.split_training_and_validation_data`.
            split = int(input_batch * (1 - kwargs["validation_split"]))
            if (
                split % self.minibatch_size != 0
                or (input_batch - split) % self.minibatch_size != 0
            ):
                raise ValidationError(
                    "Split data is not evenly divisible by minibatch size",
                    "validation_split",
                )

        # warn for synapses with n_steps=1
        # note: we don't warn if stateful, since there could be effects across runs
        if not stateful:
            target_probes = [
                p
                for p, e in zip(
                    self.model.probes,
                    getattr(self.keras_model, "_training_endpoints", []),
                )
                if not e.should_skip_target()
            ]

            synapses = [
                x.synapse is not None
                for x in (self.model.toplevel.all_connections + target_probes)
            ]
            if input_steps == 1 and self.model.toplevel is not None and any(synapses):
                warnings.warn(
                    "Running for one timestep, but the network contains "
                    "synaptic filters (which will introduce at least a "
                    "one-timestep delay); did you mean to set synapse=None?"
                )

        # set up callback to reset state after execution.
        # only necessary if simulator is stateful but this call is not stateful
        if not stateful and self.stateful:
            kwargs["callbacks"] = (kwargs.get("callbacks", None) or []) + [
                callbacks.IsolateState(self)
            ]

        # call underlying keras function
        if "predict" in func_type:
            func_args = dict(x=x, **kwargs)
        else:
            func_args = dict(x=x, y=y, **kwargs)

        outputs = getattr(self.keras_model, func_type)(**func_args)

        # update n_steps/time
        if stateful:
            self._update_steps()

        # process keras outputs
        if func_type.startswith("predict"):
            # reorganize results (will be flattened) back into dict
            if not isinstance(outputs, list):
                outputs = [outputs]
            return collections.OrderedDict(zip(self.model.probes, outputs))
        elif func_type.startswith("evaluate"):
            # return outputs as named dict
            return collections.OrderedDict(zip(self.keras_model.metrics_names, outputs))
        else:
            # return training history
            return outputs

    def step(self, **kwargs):
        """
        Run the simulation for one time step.

        Parameters
        ----------
        kwargs : dict
            See `.run_steps`

        Notes
        -----
        Progress bar is disabled by default when running via this method.
        """

        kwargs.setdefault("progress_bar", False)

        self.run_steps(1, **kwargs)

    def run(self, time_in_seconds, **kwargs):
        """
        Run the simulation for the given length of time.

        Parameters
        ----------
        time_in_seconds : float
            Run the simulator for the given number of simulated seconds.
        kwargs : dict
            See `.run_steps`
        """

        if time_in_seconds < 0:
            raise ValidationError(
                "Must be positive (got %g)" % (time_in_seconds,), attr="time_in_seconds"
            )

        steps = int(np.round(float(time_in_seconds) / self.dt))

        if steps == 0:
            warnings.warn(
                "%g results in running for 0 timesteps. Simulator "
                "still at time %g." % (time_in_seconds, self.time)
            )
        else:
            self.run_steps(steps, **kwargs)

    @require_open
    @fill_docs("stateful", data="x")
    def run_steps(self, n_steps, data=None, progress_bar=None, stateful=True):
        """
        Run the simulation for the given number of steps.

        Parameters
        ----------
        n_steps : int
            The number of simulation steps to be executed.
        data :
            {{ data }}
        progress_bar : bool
            If True, print information about the simulation status to standard
            output.
        stateful : bool
            {{ stateful }}

        Notes
        -----
        If ``unroll_simulation=x`` is specified, and ``n_steps > x``, this will
        repeatedly execute ``x`` timesteps until the the number of steps
        executed is >= ``n_steps``.
        """

        actual_steps = self.unroll * int(np.ceil(n_steps / self.unroll))

        # error checking
        if actual_steps != n_steps:
            warnings.warn(
                "Number of steps (%d) is not an even multiple of "
                "`unroll_simulation` (%d).  Simulation will run for %d steps, "
                "which may have unintended side effects."
                % (n_steps, self.unroll, actual_steps),
                RuntimeWarning,
            )

        if progress_bar is None:
            progress_bar = self.progress_bar
        progress = (
            utils.ProgressBar("Simulating", "Simulation", max_value=None)
            if progress_bar
            else utils.NullProgressBar()
        )

        with progress:
            # run the simulation
            try:
                output = self.predict_on_batch(
                    data, n_steps=actual_steps, stateful=stateful
                )
            except (tf.errors.InternalError, tf.errors.UnknownError) as e:
                if "nengo.exceptions.SimulationError" in e.message:
                    raise SimulationError(
                        "SimulationError detected; this most likely means that a "
                        "Python function (e.g. in a Node or Direct ensemble) caused "
                        "an error. See the full error log above."
                    )
                else:
                    raise e  # pragma: no cover (unknown errors)

        # update stored probe data
        for probe, val in output.items():
            if probe.sample_every is not None:
                # downsample probe according to `sample_every`
                period = probe.sample_every / self.dt
                steps = np.arange(self.n_steps - actual_steps, self.n_steps)
                val = val[:, (steps + 1) % period < 1]

            self.model.params[probe].append(val)

    def train(self, *args, **kwargs):
        """Deprecated, use `.Simulator.compile` and `.Simulator.fit` instead."""

        raise SimulationError(
            "Simulator.train has been deprecated, use Simulator.compile/fit instead"
        )

    def loss(self, *args, **kwargs):
        """Deprecated, use `.Simulator.compile` and `.Simulator.evaluate` instead."""

        raise SimulationError(
            "Simulator.loss has been deprecated, use Simulator.compile/evaluate instead"
        )

    @require_open
    @with_self
    def save_params(self, path, include_state=False, include_non_trainable=None):
        """
        Save network parameters to the given ``path``.

        Parameters
        ----------
        path : str
            Filepath of parameter output file.
        include_state : bool
            If True (default False) also save the internal simulation state.

            .. versionchanged:: 3.2.0
               Renamed from ``include_non_trainable`` to ``include_state``.

        Notes
        -----
        This function is useful for saving/loading entire models; for
        saving/loading individual objects within a model, see
        `.get_nengo_params`.
        """

        if include_non_trainable is not None:
            warnings.warn(
                "include_non_trainable is deprecated, use include_state instead",
                DeprecationWarning,
            )
            include_state = include_non_trainable

        params = list(self.keras_model.weights)
        if include_state:
            params.extend(self.tensor_graph.saved_state.values())

        np.savez_compressed(path + ".npz", *tf.keras.backend.batch_get_value(params))

        logger.info("Model parameters saved to %s.npz", path)

    @require_open
    @with_self
    def load_params(self, path, include_state=False, include_non_trainable=None):
        """
        Load network parameters from the given ``path``.

        Parameters
        ----------
        path : str
            Filepath of parameter input file.
        include_state : bool
            If True (default False) also save the internal simulation state.

            .. versionchanged:: 3.2.0
               Renamed from ``include_non_trainable`` to ``include_state``.

        Notes
        -----
        This function is useful for saving/loading entire models; for
        saving/loading individual objects within a model, see
        `.get_nengo_params`.
        """

        if include_non_trainable is not None:
            warnings.warn(
                "include_non_trainable is deprecated, use include_state instead",
                DeprecationWarning,
            )
            include_state = include_non_trainable

        params = list(self.keras_model.weights)
        if include_state:
            params.extend(self.tensor_graph.saved_state.values())

        with np.load(path + ".npz") as vals:
            if len(params) != len(vals.files):
                raise SimulationError(
                    "Number of saved parameters in %s (%d) != number of variables in "
                    "the model (%d)" % (path, len(vals.files), len(params))
                )
            tf.keras.backend.batch_set_value(
                zip(params, (vals["arr_%d" % i] for i in range(len(vals.files))))
            )

        logger.info("Model parameters loaded from %s.npz", path)

    @require_open
    def freeze_params(self, objs):
        """
        Stores the live parameter values from the simulation back into a
        Nengo object definition.

        This can be helpful for reusing a NengoDL model inside a different
        Simulator.  For example:

        .. testcode::

            with nengo.Network() as net:
                ens = nengo.Ensemble(10, 1)

            with nengo_dl.Simulator(net) as sim:
                # < run some optimization >
                sim.freeze_params(net)

            with nengo.Simulator(net) as sim2:
                # run the network in the default Nengo simulator, with the
                # trained parameters
                sim2.run(1.0)

        .. testoutput::
            :hide:

            ...

        Parameters
        ----------
        obj : (list of) ``NengoObject``
            The Nengo object(s) into which parameter values will be stored.
            Note that these objects must be members of the Network used to
            initialize the Simulator.

        Notes
        -----
        This modifies the source object in-place, and it may slightly modify
        the structure of that object.  The goal is to have the object produce
        the same output as it would if run in the NengoDL simulator.  It may
        not be possible to accurately freeze all possible object; if you run
        into errors in this process, try manually extracting the parameters you
        need in your model (from ``sim.data``).
        """

        if not isinstance(objs, (list, tuple)):
            objs = [objs]

        for obj in objs:
            if obj not in [self.model.toplevel] + self.model.toplevel.all_objects:
                raise ValueError(
                    "%s is not a member of the Network used to "
                    "initialize the Simulator"
                )

            if not isinstance(obj, (Network, Ensemble, Connection)):
                raise TypeError(
                    "Objects of type %s do not have parameters to store" % type(obj)
                )

            if isinstance(obj, Network):
                todo = obj.all_ensembles + obj.all_connections
            else:
                todo = [obj]

            for o, params in zip(todo, self.get_nengo_params(todo)):
                for k, v in params.items():
                    setattr(o, k, v)

    def get_nengo_params(self, nengo_objs, as_dict=False):
        """
        Extract model parameters in a form that can be used to initialize
        Nengo objects in a different model.

        For example:

        .. testcode::

            with nengo.Network() as net:
                a = nengo.Ensemble(10, 1)
                b = nengo.Ensemble(10, 1)
                c = nengo.Connection(a, b)

            with nengo_dl.Simulator(net) as sim:
                # < do some optimization >
                params = sim.get_nengo_params([a, b, c])

            with nengo.Network() as new_net:
                # < build some other network >

                # now we want to insert two connected ensembles with
                # the same parameters as our previous network:
                d = nengo.Ensemble(10, 1, **params[0])
                e = nengo.Ensemble(10, 1, **params[1])
                f = nengo.Connection(d, e, **params[2])

        Parameters
        ----------
        nengo_objs : (list of) `~nengo.Ensemble` or `~nengo.Connection`
            A single object or list of objects for which we want to get the
            parameters.
        as_dict : bool
            If True, return the values as a dictionary keyed by object label,
            instead of a list (the default).  Note that in this case labels
            must be unique.

        Returns
        -------
        params : (list or dict) of dicts
            kwarg dicts corresponding to ``nengo_objs`` (passing these
            dicts as kwargs when creating new Nengo objects will result in a
            new object with the same parameters as the source object).  A
            single kwarg dict if a single object was passed in, or a list
            (dict if ``as_dict=True``) of kwargs corresponding to multiple
            input objects.
        """

        if isinstance(nengo_objs, (list, tuple)):
            scalar = False
        else:
            scalar = True
            nengo_objs = [nengo_objs]

        # convert neurons to the parent ensemble
        nengo_objs = [
            obj.ensemble if isinstance(obj, Neurons) else obj for obj in nengo_objs
        ]

        # find all the data we need to fetch
        fetches = []
        for obj in nengo_objs:
            if isinstance(obj, Connection):
                if compat.conn_has_weights(obj):
                    fetches.append((obj, "weights"))
            elif isinstance(obj, Ensemble):
                if isinstance(obj.neuron_type, Direct):
                    # we cannot transfer direct ensemble parameters, because
                    # the nengo builder ignores the encoders specified for
                    # a direct ensemble
                    raise ValueError(
                        "get_nengo_params will not work correctly for "
                        "Direct neuron ensembles. Try manually translating "
                        "your network using `sim.data` instead."
                    )

                fetches.extend([(obj, "scaled_encoders"), (obj, "bias")])
            else:
                raise ValueError(
                    "Can only get Nengo parameters for Ensembles or Connections"
                )

        # get parameter values from simulation
        data = self.data.get_params(*fetches)

        # store parameter values in a form that can be loaded in nengo
        params = []
        idx = 0
        for obj in nengo_objs:
            if isinstance(obj, Connection):
                if not compat.conn_has_weights(obj):
                    params.append({"transform": None})
                    continue

                weights = data[idx]
                idx += 1
                if isinstance(obj.pre_obj, Ensemble):
                    params.append(
                        {
                            "solver": NoSolver(weights.T, weights=False),
                            "function": lambda x, weights=weights: np.zeros(
                                weights.shape[0]
                            ),
                            "transform": compat.default_transform,
                        }
                    )
                elif isinstance(obj.transform, Convolution):
                    transform = copy.copy(obj.transform)
                    # manually bypass the read-only check (we are sure that
                    # nothing else has a handle to the new transform at this
                    # point, so this won't cause any problems)
                    Convolution.init.data[transform] = weights
                    params.append({"transform": transform})
                else:
                    if all(x == 1 for x in weights.shape):
                        weights = np.squeeze(weights)
                    params.append({"transform": weights})
            else:
                # note: we don't want to change the original gain (even though
                # it is rolled into the encoder values), because connections
                # direct to `ens.neurons` will still use the gains (and those
                # gains are not updated during training, only the encoders)
                gain = self.model.params[obj].gain

                # the encoders we get from the simulation are the actual
                # weights we want in the simulation. but during the build
                # process, gains and radius will be applied to the encoders.
                # so we need to undo that scaling here, so that the build
                # process will result in the correct values.
                encoders = data[idx] * obj.radius / gain[:, None]

                params.append(
                    {
                        "encoders": encoders,
                        "normalize_encoders": False,
                        "gain": gain,
                        "bias": data[idx + 1],
                        "max_rates": Ensemble.max_rates.default,
                        "intercepts": Ensemble.intercepts.default,
                    }
                )
                idx += 2

        # return params in appropriate format
        if scalar:
            return params[0]

        if as_dict:
            param_dict = {}
            for obj, p in zip(nengo_objs, params):
                if obj.label in param_dict:
                    raise ValueError(
                        "Duplicate label ('%s') detected; cannot return "
                        "parameters with as_dict=True" % obj.label
                    )
                else:
                    param_dict[obj.label] = p
            params = param_dict

        return params

    @require_open
    @with_self
    def check_gradients(self, inputs=None, outputs=None, atol=1e-5, rtol=1e-3):
        """
        Perform gradient checks for the network (used to verify that the
        analytic gradients are correct).

        Raises a simulation error if the difference between analytic and
        numeric gradient is greater than ``atol + rtol * numeric_grad``
        (elementwise).

        Parameters
        ----------
        inputs : list of `numpy.ndarray`
            Input values for all the input Nodes in the model (ordered according to
            the order in which Nodes were added to the model). If None, will use all
            zeros.
        outputs : list of `~nengo.Probe`
            Compute gradients wrt this output (if None, computes wrt each
            output probe).
        atol : float
            Absolute error tolerance.
        rtol : float
            Relative (to numeric grad) error tolerance.

        Notes
        -----
        Calling this method will reset the internal simulation state.
        """

        if self.tensor_graph.inference_only:
            raise SimulationError(
                "Network was created with inference_only=True, cannot "
                "compute gradients"
            )

        if inputs is None:
            n_steps = self.unroll * 2
            inputs = [
                np.zeros(
                    tuple(n_steps if s is None else s for s in x.shape),
                    x.dtype.as_numpy_dtype(),
                )
                for x in self.keras_model.inputs[:-1]
            ]
        else:
            n_steps = inputs[0].shape[1]

        if outputs is None:
            outputs = self.model.probes

        # compute_gradients expects to be called with a function that works in
        # specific ways, so we wrap the model to work in the way it expects
        @tf.function
        @tf.autograph.experimental.do_not_convert
        def arg_func(*args, output=None):
            for i, x in enumerate(args):
                x.set_shape(inputs[i].shape)

            args += (tf.ones((self.minibatch_size, 1), dtype=np.int32) * n_steps,)

            out = self.tensor_graph(args, training=True)
            self.tensor_graph.build_post()

            if self.stateful:
                # reset state
                for var in self.tensor_graph.saved_state.values():
                    var.assign(var.initial_value)

            # drop steps_run
            out = out[:-1]

            # get selected output
            out = out[self.model.probes.index(output)]

            return out

        self.reset(
            include_probes=False, include_trainable=False, include_processes=False
        )

        grads = dict()
        for output in outputs:
            # TODO: switch to eager evaluation
            with tf.compat.v1.keras.backend.get_session().as_default():
                analytic, numeric = tf.test.compute_gradient(
                    partial(arg_func, output=output), inputs
                )
            grads[output] = dict()
            grads[output]["analytic"] = analytic
            grads[output]["numeric"] = numeric

            for a, n in zip(analytic, numeric):
                if np.any(np.isnan(a)) or np.any(np.isnan(n)):
                    raise SimulationError("NaNs detected in gradient")
                fail = abs(a - n) >= atol + rtol * abs(n)
                if np.any(fail):
                    raise SimulationError(
                        "Gradient check failed\n"
                        "numeric values:\n%s\n"
                        "analytic values:\n%s\n" % (n[fail], a[fail])
                    )

        logger.info("Gradient check passed")

        return grads

    def trange(self, sample_every=None, dt=None):
        """
        Create a vector of simulation step times matching probed data.

        Note that the range does not start at 0 as one might expect, but at
        the first timestep (i.e., ``dt``).

        Parameters
        ----------
        sample_every : float (Default: None)
            The sampling period of the probe to create a range for.
            If None, a time value for every ``dt`` will be produced.
        """

        if dt is not None:
            if sample_every is not None:
                raise ValidationError(
                    "Cannot specify both `dt` and `sample_every`. "
                    "Use `sample_every` only.",
                    attr="dt",
                    obj=self,
                )
            warnings.warn(
                "`dt` is deprecated. Use `sample_every` instead.", DeprecationWarning
            )
            sample_every = dt

        period = 1 if sample_every is None else sample_every / self.dt
        steps = np.arange(1, self.n_steps + 1)
        return self.dt * steps[steps % period < 1]

    def close(self):
        """
        Close the simulation, freeing resources.

        Notes
        -----
        The simulation cannot be restarted after it is closed.
        """

        if not self.closed:
            self.keras_model = None
            self.tensor_graph = None
            self.graph = None
            self._closed_attrs = ["keras_model", "tensor_graph", "graph"]

            self.closed = True

    def get_name(self, obj):
        """
        Returns the standardized string name for input Nodes or output Probes.

        These are used when referring to inputs/outputs by string in Keras.

        Parameters
        ----------
        obj : `nengo.Node` or `nengo.Probe`
            Input Node or output Probe

        Returns
        -------
        name : str
            Name of the given object
        """

        if isinstance(obj, Node):
            if obj not in self.node_inputs:
                raise ValidationError(
                    "%s is not an input Node (a nengo.Node with "
                    "size_in==0), or is from a different network." % obj,
                    "obj",
                )
        elif isinstance(obj, Probe):
            if obj not in self.tensor_graph.probe_arrays:
                raise ValidationError("%s is from a different network." % obj, "obj")
        else:
            raise ValidationError(
                "%s is of an unknown type (%s); should be nengo.Node "
                "or nengo.Probe" % (obj, type(obj)),
                "obj",
            )
        return self.tensor_graph.io_names[obj]

    def _standardize_data(self, data, objects, broadcast_unary=False):
        """
        Converts data to the standardized input format (named string dicts).

        Parameters
        ----------
        data : `numpy.ndarray` or list or dict
            Input data in one of the formats supported by fit/predict/eval.
        objects : list of `nengo.Node` or `nengo.Probe`
            List of input Nodes or output Probes in the model (depending on which
            kind of data is being standardized).
        broadcast_unary: bool
            If True, singular (e.g. non-list/dict) inputs will be applied to all
            ``objects``, otherwise will only be applied to first item in ``objects``.

        Returns
        -------
        data : dict of {str: object}
            Elements of data reorganized into standardized data structure (named
            string dict).
        """

        if data is None:
            return data

        if not isinstance(data, (list, tuple, dict)):
            # convert unary inputs to length-1 lists
            data = [data]
            if broadcast_unary:
                data *= len(objects)

        if isinstance(data, (list, tuple)):
            if len(data) != len(objects):
                warnings.warn(
                    "Number of data values (%d) does not match number of %ss (%d); "
                    "it is recommended to use an explicit input dictionary in this "
                    "case, so that the assignment of data to objects is unambiguous."
                    % (len(data), type(objects[0]).__name__, len(objects))
                )

            # convert list to named dict
            data = collections.OrderedDict(
                (self.get_name(obj), val) for obj, val in zip(objects, data)
            )
        elif isinstance(data, dict):
            # convert objects to string names
            data = collections.OrderedDict(
                (obj if isinstance(obj, str) else self.get_name(obj), val)
                for obj, val in data.items()
            )

        return data

    def _generate_inputs(self, data=None, n_steps=None):
        """
        Generate inputs for the network (the output values of each Node with
        no incoming connections).

        Parameters
        ----------
        data : list or dict of {`~nengo.Node` or str: `~numpy.ndarray`}
            Override the values of input Nodes with the given data.  Arrays
            should have shape ``(sim.minibatch_size, n_steps, node.size_out)``.
        n_steps : int
            Number of simulation timesteps for which to generate input data

        Returns
        -------
        node_vals : dict of {str: `~numpy.ndarray}
            Simulation values for all the input Nodes in the network.
        """

        if data is None:
            data = {}

        if not isinstance(data, (list, tuple, dict, np.ndarray, tf.Tensor)):
            # data is some kind of generator, so we don't try to modify it (too many
            # different types of generators this could be)
            if n_steps is not None:
                raise SimulationError(
                    "Cannot automatically add n_steps to generator with type %s; "
                    "please specify n_steps manually as the first element in the "
                    "values yielded from generator, remembering that it needs to "
                    "be repeated to have shape (batch_size, 1)" % type(data)
                )

            return data

        data = self._standardize_data(data, list(self.node_inputs.keys()))

        if len(data) == 0:
            data_batch = data_steps = None
        else:
            data_batch, data_steps = next(iter(data.values())).shape[:2]

        batch_size = self.minibatch_size if data_batch is None else data_batch
        if n_steps is None:
            if data_steps is None:
                raise ValidationError(
                    "Must specify either input data or n_steps", "data"
                )
            n_steps = data_steps

        input_vals = collections.OrderedDict()

        # fill in data for input nodes
        for node, output in self.tensor_graph.input_funcs.items():
            name = self.get_name(node)

            if name in data:
                node_val = data[name]
            elif isinstance(output, np.ndarray):
                # tile to n_steps/minibatch size
                node_val = np.tile(output[None, None, :], (batch_size, n_steps, 1))
            else:
                # call output function to determine value
                node_val = np.zeros(
                    (batch_size, n_steps, node.size_out),
                    dtype=np.dtype(self.tensor_graph.dtype),
                )

                for i in range(n_steps):
                    # note: need to copy the output of func, as func
                    # may mutate its outputs in-place on subsequent calls.
                    # this assignment will broadcast the output along the
                    # minibatch dimension if required.
                    # note: we still call the function even if the output
                    # is not being used in the graph, because it may have side-effects
                    node_val[:, i] = [
                        func((i + self.n_steps + 1) * self.dt) for func in output
                    ]

            input_vals[name] = node_val

        for name in data:
            if name not in input_vals:
                raise ValidationError(
                    "Input contained entry for '%s', which is not a valid input name"
                    % name,
                    "data",
                )

        # fill in n_steps
        input_vals["n_steps"] = np.resize(n_steps, (batch_size, 1)).astype(np.int32)

        return input_vals

    def _check_data(self, data, batch_size=None, n_steps=None, nodes=True):
        """
        Performs error checking on simulation data.

        Parameters
        ----------
        data : dict of {str: `~numpy.ndarray` or ``tf.Tensor``}
            Array of data associated with given objects in model (Nodes or
            Probes)
        batch_size : int
            Number of elements in batch (if None, will just verify that all
            data items have same batch size)
        n_steps : int
            Number of simulation steps (if None, will just verify that all
            data items have same number of steps)
        nodes : bool
            If True the data being validated is associated with Nodes, if False the
            data is associated with Probes.

        Notes
        -----
        This may modify ``data`` in-place, if it contains data that is not evenly
        divisible by ``Simulator.minibatch_size``.
        """

        if not isinstance(data, dict):
            # data is a generator, so don't perform validation
            return

        # make sure data is evenly divisible by minibatch size
        for k, v in data.items():
            try:
                data_batch = v.shape[0]
            except IndexError:
                # v is a scalar
                continue

            if (
                data_batch > self.minibatch_size
                and data_batch % self.minibatch_size != 0
            ):
                warnings.warn(
                    "Number of elements in input data (%d) is not evenly divisible by "
                    "Simulator.minibatch_size (%d); input data will be truncated."
                    % (data_batch, self.minibatch_size)
                )
                data_batch -= data_batch % self.minibatch_size
                data[k] = v[:data_batch]

        # exclude n_steps from normal data checking
        data_n_steps = data.get("n_steps", None)
        data = {k: val for k, val in data.items() if k != "n_steps"}

        for name, x in data.items():
            # check that name is valid
            if nodes:
                valid_names = [self.get_name(n) for n in self.node_inputs]
                if name not in valid_names:
                    raise ValidationError(
                        "'%s' is not a valid node name; perhaps the name is wrong (it "
                        "should match the `label` on the Node), or this is not an "
                        "input Node (a Node with size_in==0) in this network. "
                        "Valid names are: %s." % (name, valid_names),
                        "data",
                    )
            else:
                valid_names = [self.get_name(p) for p in self.model.probes]
                if name not in valid_names:
                    raise ValidationError(
                        "'%s' is not a valid probe name; perhaps the name is wrong (it "
                        "should match the `label` on the Probe), or this is not a "
                        "Probe in this network. Valid names are: %s."
                        % (name, valid_names),
                        "data",
                    )

            # generic shape checks
            if len(x.shape) != 3:
                raise ValidationError(
                    "should have rank 3 (batch_size, n_steps, dimensions), "
                    "found rank %d" % len(x.shape),
                    "%s data" % name,
                )
            if x.shape[0] < self.minibatch_size:
                raise ValidationError(
                    "Batch size of data (%d) less than Simulator `minibatch_size` (%d)"
                    % (x.shape[0], self.minibatch_size),
                    "%s data" % name,
                )
            if nodes and x.shape[1] % self.unroll != 0:
                raise ValidationError(
                    "The number of timesteps in input data (%s) must be evenly "
                    "divisible by unroll_simulation (%s)" % (x.shape[1], self.unroll),
                    "data",
                )

        # check that shapes match the given values (if specified) or are
        # internally consistent (if not)
        args = [batch_size, n_steps]
        labels = ["batch size", "number of timesteps"]

        for i in range(2):
            if args[i] is None:
                if i == 1 and not nodes:
                    # we don't apply this check to probes, because target values can
                    # have different values for n_steps (as long as it matches what is
                    # expected by the loss function)
                    continue

                if len(data) > 0:
                    val = next(iter(data.values())).shape[i]
                for n, x in data.items():
                    if x.shape[i] != val:
                        raise ValidationError(
                            "Elements have different %s: %s vs %s"
                            % (labels[i], val, x.shape[i]),
                            "data",
                        )
            else:
                for n, x in data.items():
                    if x.shape[i] != args[i]:
                        raise ValidationError(
                            "Data for %s has %s=%s, which does not match "
                            "expected size (%s)" % (n, labels[i], x.shape[i], args[i]),
                            "data",
                        )

        if (
            n_steps is not None
            and not self.tensor_graph.use_loop
            and n_steps != self.unroll
        ):
            raise ValidationError(
                "When use_loop=False, n_steps (%d) must exactly match "
                "unroll_simulation (%d)" % (n_steps, self.unroll),
                "n_steps",
            )

        if nodes:
            # validate special n_steps input

            if data_n_steps is None:
                raise ValidationError("Must specify 'n_steps' in input data", "data")
            if (
                batch_size is None
                and (data_n_steps.ndim != 2 or data_n_steps.shape[1] != 1)
            ) or (batch_size is not None and data_n_steps.shape != (batch_size, 1)):
                raise ValidationError(
                    "'n_steps' has wrong shape; should be %s (note that this is just "
                    "the integer n_steps value repeated)" % ((batch_size, 1),),
                    "data",
                )
            if not np.all(data_n_steps == data_n_steps[0, 0]):
                raise ValidationError(
                    "'n_steps' should all have the same value", "data"
                )
            if n_steps is not None and not np.all(data_n_steps == n_steps):
                raise ValidationError(
                    "`n_steps` input does not match the requested number of steps",
                    "data",
                )

    @with_self
    def _update_steps(self):
        if not hasattr(self, "_step_tensor"):
            # cache this so we aren't adding new ops every time we call this function
            self._step_tensor = self.tensor_graph.get_tensor(self.model.step)

        self._n_steps = tf.keras.backend.get_value(self._step_tensor).item()
        self._time = self._n_steps * self.dt

    @property
    def dt(self):
        """The time (in seconds) represented by one simulation timestep."""
        return self.model.dt

    @dt.setter
    def dt(self, _):
        raise ReadonlyError(attr="dt", obj=self)

    @property
    def n_steps(self):
        """The current simulation timestep."""
        return self._n_steps

    @property
    def time(self):
        """The current simulation time."""
        return self._time

    @property
    def seed(self):
        """The simulation random seed."""
        return self.tensor_graph.seed

    @require_open
    def __enter__(self):
        self._graph_context = self.graph.as_default()
        self._device_context = self.graph.device(self.tensor_graph.device)

        self._graph_context.__enter__()
        self._device_context.__enter__()

        self._keras_dtype = tf.keras.backend.floatx()
        tf.keras.backend.set_floatx(self.tensor_graph.dtype)

        return self

    @require_open
    def __exit__(self, *args):
        tf.keras.backend.set_floatx(self._keras_dtype)

        self._device_context.__exit__(*args)
        self._graph_context.__exit__(*args)

        self.close()

    def __del__(self):
        """
        Raise a RuntimeWarning if the Simulator is deallocated while open.
        """

        if self.closed is not None and not self.closed:
            warnings.warn(
                "Simulator with model=%s was deallocated while open. "
                "Simulators should be closed manually to ensure resources "
                "are properly freed." % self.model,
                RuntimeWarning,
            )
            self.close()

    def __getstate__(self):
        raise NotImplementedError(
            "TensorFlow does not support pickling; see "
            "https://www.nengo.ai/nengo-dl/simulator.html"
            "#saving-and-loading-parameters "
            "for information on how to save/load a NengoDL model."
        )

    def __getattribute__(self, name):
        if super().__getattribute__("closed") and name in super().__getattribute__(
            "_closed_attrs"
        ):
            raise SimulatorClosed(
                "Cannot access Simulator.%s after Simulator is closed" % name
            )

        return super().__getattribute__(name)


class SimulationData(collections.Mapping):
    """
    Data structure used to access simulation data from the model.

    The main use case for this is to access Probe data; for example,
    ``probe_data = sim.data[my_probe]``.  However, it is also
    used to access the parameters of objects in the model; for example, after
    the model has been optimized via `.Simulator.fit`, the updated
    encoder values for an ensemble can be accessed via
    ``trained_encoders = sim.data[my_ens].encoders``.

    Parameters
    ----------
    sim : `.Simulator`
        The simulator from which data will be drawn
    minibatched : bool
        If False, discard the minibatch dimension on probe data

    Notes
    -----
    SimulationData shouldn't be created/accessed directly by the user, but
    rather via ``sim.data`` (which is an instance of SimulationData).
    """

    def __init__(self, sim, minibatched):
        self.sim = sim
        self.minibatched = minibatched

    def __getitem__(self, obj):
        """Return the data associated with ``obj``.

        Parameters
        ----------
        obj : `~nengo.Probe` or `~nengo.Ensemble` or `~nengo.Connection`
            Object whose simulation data is being accessed

        Returns
        -------
        data : `~numpy.ndarray` or \
               `~nengo.builder.ensemble.BuiltEnsemble` or \
               `~nengo.builder.connection.BuiltConnection`
            Array containing probed data if ``obj`` is a
            `~nengo.Probe`, otherwise the corresponding
            parameter object
        """

        if obj not in self.sim.model.params:
            raise ValidationError(
                "Object is not in parameters of model %s" % self.sim.model, str(obj)
            )

        data = self.sim.model.params[obj]

        if isinstance(obj, Probe):
            if len(data) == 0:
                return []
            data = np.concatenate(data, axis=1)
            if not self.minibatched:
                data = data[0]

            data.setflags(write=False)
        elif isinstance(obj, Ensemble):
            if isinstance(obj.neuron_type, Direct):
                # direct mode ensemble
                gain = bias = None
                scaled_encoders = encoders = self.get_params((obj, "scaled_encoders"))[
                    0
                ]
            else:
                # get the live simulation values
                scaled_encoders, bias = self.get_params(
                    (obj, "scaled_encoders"), (obj, "bias")
                )

                # infer the related values (rolled into scaled_encoders)
                gain = (
                    obj.radius
                    * np.linalg.norm(scaled_encoders, axis=-1)
                    / np.linalg.norm(data.encoders, axis=-1)
                )
                encoders = obj.radius * scaled_encoders / gain[:, None]

            # figure out max_rates/intercepts from neuron model
            max_rates, intercepts = obj.neuron_type.max_rates_intercepts(gain, bias)

            data = BuiltEnsemble(
                data.eval_points,
                encoders,
                intercepts,
                max_rates,
                scaled_encoders,
                gain,
                bias,
            )
        elif isinstance(obj, Connection):
            # get the live simulation values
            weights = (
                self.get_params((obj, "weights"))[0]
                if compat.conn_has_weights(obj)
                else None
            )

            # impossible to recover transform
            transform = None

            data = BuiltConnection(
                data.eval_points, data.solver_info, weights, transform
            )

        return data

    def get_params(self, *obj_attrs):
        """
        Returns the current parameter values for the given objects.

        Parameters
        ----------
        obj_attrs : list of (``NengoObject``, str)
            The Nengo object and attribute of that object for which we want
            to know the parameter values (each object-attribute pair specified
            as a tuple argument to the function).

        Returns
        -------
        params : list of `~numpy.ndarray`
            Current values of the requested parameters

        Notes
        -----
        Parameter values should be accessed through ``sim.data[my_obj]``
        (which will call this function if necessary), rather than directly
        through this function.
        """

        if self.sim.closed:
            warnings.warn(
                "Checking parameters after simulator is closed; "
                "cannot fetch live values, so the initial values "
                "will be returned."
            )

            return [
                getattr(self.sim.model.params[obj], attr) for obj, attr in obj_attrs
            ]

        params = []
        sigs = []
        fetches = {}
        for obj, attr in obj_attrs:

            sig_obj, sig_attr = self._attr_map(obj, attr)
            sig = self.sim.model.sig[sig_obj][sig_attr]
            sigs.append(sig)

            if sig not in self.sim.tensor_graph.signals:
                # if sig isn't in sig_map then that means it isn't used
                # anywhere in the simulation (and therefore never changes), so
                # we can safely return the static build value
                params.append(getattr(self.sim.model.params[obj], attr))
            else:
                # this is a live parameter value we need to fetch from the
                # simulation. we queue them up and fetch them all at once to
                # be more efficient
                placeholder = object()
                fetches[placeholder] = self.sim.tensor_graph.get_tensor(sig)
                params.append(placeholder)

        # get the live parameter values
        fetched = dict(
            zip(
                fetches.keys(), tf.keras.backend.batch_get_value(list(fetches.values()))
            )
        )

        # final updating of parameters
        for i, sig in enumerate(sigs):
            # fill in placeholder values
            if type(params[i]) == object:
                params[i] = fetched[params[i]]

            if sig.minibatched and not self.minibatched:
                # drop minibatch dimension
                params[i] = params[i][0]

        return params

    def _attr_map(self, obj, attr):
        """
        Maps from ``sim.data[obj].attr`` to the equivalent
        ``model.sig[obj][attr]``.

        Parameters
        ----------
        obj : ``NengoObject``
            The nengo object for which we want to know the parameters
        attr : str
            The parameter of ``obj`` to be returned

        Returns
        -------
        obj : ``NengoObject``
            The nengo object to key into ``model.sig``
        attr : str
            The name of the signal corresponding to input attr

        """

        if isinstance(obj, Ensemble) and attr == "bias":
            return obj.neurons, attr
        elif isinstance(obj, Ensemble) and attr == "scaled_encoders":
            return obj, "encoders"

        return obj, attr

    def __len__(self):
        return len(self.sim.model.params)

    def __iter__(self):
        return iter(self.sim.model.params)
